# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import importlib.util
import json
import os

import numpy as np
import torch

from .._quant_common.helper_modules import *
from .._quant_common.quant_config import get_hqt_config
from ..utils.logger import logger

deepspeed_exists = False
if importlib.util.find_spec("deepspeed"):  # check if deepspeed is installed
    deepspeed_exists = True

UNMEASURED_MODELS = "UnmeasuredModels"


class ModuleInfo:
    def __init__(self, type, patched_module):
        self.type = type
        self.patched_module = patched_module


class ModuleConfig:
    def __init__(self, inputs=(None,), outputs=(None,), params=None):
        self.inputs = inputs
        self.outputs = outputs
        self.params = params if params is not None else {}


class ModuleExtraConfig:
    def __init__(self, inputs=(None,), outputs=(None,), params=None, scale=None, config_params=None):
        self.inputs = inputs
        self.outputs = outputs
        self.params = params if params is not None else {}
        self.scale = scale
        self.config_params = config_params if config_params is not None else {}


class ModuleType:
    def __init__(self, num_inputs, param_names, num_outputs, required_output):
        self.num_inputs = num_inputs
        self.param_names = param_names
        self.num_outputs = num_outputs
        self.required_output = required_output


mod_types = {
    "linear": ModuleType(1, ["weight"], 1, False),
    "matmul": ModuleType(2, [], 1, False),
    "kv_cache": ModuleType(1, [], 1, False),
    "softmax": ModuleType(1, [], 1, True),
    "fused_sdpa": ModuleType(3, [], 2, True),
}
descale_fcn = lambda x, scale: torch.mul(x, scale)
scale_fcn = lambda x, scale: torch.div(x, scale)
cast_fcn = lambda x, dtype: x.to(dtype=dtype)
cast_to_fp8_fcn = lambda x, dtype, scale_inv=None: torch.ops.hpu.cast_to_fp8_v2(x, scale_inv, False, False, dtype)[0]
cast_from_fp8_fcn = lambda x, dtype, scale=None: torch.ops.hpu.cast_from_fp8(x, scale, dtype)


class ShapeList:
    data = None


def rec_fn(x, fn):
    if isinstance(x, dict):
        return {k: rec_fn(x[k], fn) for k in x}
    elif isinstance(x, list):
        return [rec_fn(k, fn) for k in x]
    elif isinstance(x, tuple):
        return tuple([rec_fn(k, fn) for k in x])
    else:
        return fn(x)


def save_json(d, fname):
    with open(fname, "w") as f:
        json.dump(d, f, indent=4)


def load_json(fname):
    with open(fname, "r") as f:
        d = json.load(f)
    return d


def save_npz(d, fname):
    np.savez(fname, d)


def load_npz(fname):
    d = np.load(fname, allow_pickle=True)
    return d["arr_0"].item()


def save_file(model, d, source_format, fname, mode):
    config = get_hqt_config(model)
    logger.debug("Saving %s file: %s", mode, fname)
    ext = os.path.splitext(fname)[1]
    target_format = file_functions[ext][0]
    dc = rec_fn(d, format_functions[(source_format, target_format)])
    df = {
        "GlobalRank": config.cfg["global_rank"],
        "LocalRank": config.cfg["local_rank"],
        "Mode": mode,
        "Nodes": dc,
    }
    try:
        file_functions[ext][1](df, fname)
    except:
        pass


# convert module config data to other format
def module_convert(m, fcn):
    mt = ModuleConfig(
        tuple([fcn(x) for x in m.inputs]),
        (
            tuple(
                [fcn(m.outputs)],
            )
            if type(m.outputs) == np.ndarray
            else tuple([fcn(y) for y in m.outputs])
        ),
        {k: fcn(m.params[k]) for k in m.params},
    )
    return mt


def fix_fields(d):
    if "input" in d:
        d["inputs"] = d.pop("input")
    if "output" in d:
        d["outputs"] = d.pop("output")
    return d


def load_file(fname, target_format, fail_on_file_not_exist):
    logger.debug("Loading file: %s", fname)
    ext = os.path.splitext(fname)[1]
    source_format = file_functions[ext][0]
    d = {}
    if os.path.isfile(fname):
        d = file_functions[ext][2](fname)
    elif fail_on_file_not_exist:
        raise FileNotFoundError(f"Failed to load file {fname}")
    if "Nodes" in d:
        dc = {k: ModuleConfig(**fix_fields(d["Nodes"][k])) for k in d["Nodes"]}
        dc = {k: module_convert(dc[k], format_functions[(source_format, target_format)]) for k in dc}
    else:
        dc = {}
    return dc


def save_scales(model, d, source_format, fname):
    dc = {k: d[k].__dict__ for k in d}
    save_file(model, dc, source_format, fname, "Scale")


def load_scales(fname, target_format):
    logger.debug("Loading scales file %s", fname)
    d = load_file(fname, target_format, False)
    return d


def convert_scales_to_tensors_dict(scales_obj, scales_file_format, hp_dtype):
    scales_temp = {k: scales_obj[k].__dict__ for k in scales_obj}
    scales_temp = format_functions_rec((scales_file_format, torch.Tensor))(scales_temp)
    scales_temp = rec_fn(scales_temp, lambda x: x.to(dtype=hp_dtype, device="hpu"))
    scales = {k: ModuleConfig(**scales_temp[k]) for k in scales_temp}
    return scales


file_functions = {
    ".json": (list, save_json, load_json),
    ".npz": (np.ndarray, save_npz, load_npz),
}

format_functions = {
    (torch.Tensor, torch.Tensor): lambda x: x,
    (np.ndarray, np.ndarray): lambda x: x,
    (list, list): lambda x: x,
    (torch.Tensor, np.ndarray): lambda x: x.detach().cpu().float().numpy(),
    (torch.Tensor, list): lambda x: x.detach().cpu().float().numpy().tolist(),
    (np.ndarray, torch.Tensor): torch.tensor,
    (np.ndarray, list): lambda x: x.tolist(),
    (list, torch.Tensor): torch.tensor,
    (list, np.ndarray): lambda x: np.array(x),
    (list, ShapeList): lambda x: [int(s) for s in x[0]],
}


format_functions_rec = lambda k: functools.partial(rec_fn, fn=format_functions[k])

mod_default_dict = {
    "Matmul": ModuleInfo("matmul", PatchedMatmul),
    "Linear": ModuleInfo("linear", PatchedLinear),
    "RowParallelLinear": ModuleInfo("linear", PatchedRowParallelLinear),
    "ColumnParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear),
    "MergedColumnParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear),
    "QKVParallelLinear": ModuleInfo("linear", PatchedColumnParallelLinear),
    "FalconLinear": ModuleInfo("linear", PatchedLinear),
    "KVCache": ModuleInfo("kv_cache", PatchedKVCache),
    "VLLMKVCache": ModuleInfo("kv_cache", PatchedVLLMKVCache),
    "Conv2d": ModuleInfo("linear", PatchedConv2d),
    "LoRACompatibleLinear": ModuleInfo("linear", PatchedLoRACompatibleLinear),
    "LoRACompatibleConv": ModuleInfo("linear", PatchedLoRACompatibleConv),
    "Softmax": ModuleInfo("softmax", PatchedSoftmax),
    "ModuleFusedSDPA": ModuleInfo("fused_sdpa", PatchedModuleFusedSDPA),
}


if deepspeed_exists:
    mod_default_dict.update(
        {
            "LinearLayer": ModuleInfo("linear", PatchedLinear),
            "LinearAllreduce": ModuleInfo("linear", PatchedLinearAllReduce),
            "ScopedLinearAllReduce": ModuleInfo("linear", PatchedLinearAllReduce),
            "LmHeadLinearAllreduce": ModuleInfo("linear", PatchedLmHeadLinearAllreduce),
        }
    )


class ModInstInfo:
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent


parent_child_mod_dict = {}


def generate_model_info(model):
    def create_mod_info_recursion(parent):
        for name, mod in parent.named_children():
            parent_child_mod_dict[mod] = ModInstInfo(name, parent)
            create_mod_info_recursion(mod)

    create_mod_info_recursion(model)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import habana_frameworks.torch.core as htcore
import habana_frameworks.torch.utils.experimental as htexp
import torch

from .common import *

GAUDI2 = htexp.synDeviceType.synDeviceGaudi2
GAUDI3 = htexp.synDeviceType.synDeviceGaudi3

EXP_WIDTH = {
    torch.float32: 8,
    torch.bfloat16: 8,
    torch.float8_e4m3fn: 4,
    torch.float8_e5m2: 5,
}


def get_default_exp_bias(dtype):
    exp_width = EXP_WIDTH[dtype]
    return 2 ** (exp_width - 1) - 1


EXP_BIAS_SETS = {
    (GAUDI2, torch.float8_e4m3fn): [3, 7, 11, 15],
    (GAUDI2, torch.float8_e5m2): [15],
    (GAUDI3, torch.float8_e4m3fn): range(0, 63),
    (GAUDI3, torch.float8_e5m2): range(0, 63),
}

MAX_RANGE = {
    torch.float32: 2 ** ((2**8 - 2 - get_default_exp_bias(torch.float32))) * (2 - 2 ** -(23)),
    torch.bfloat16: 2 ** ((2**8 - 2 - get_default_exp_bias(torch.bfloat16))) * (2 - 2 ** -(7)),
    torch.float8_e4m3fn: 2 ** ((2**4 - 2 - get_default_exp_bias(torch.float8_e4m3fn))) * (2 - 2 ** -(8 - 1 - 4)),
    torch.float8_e5m2: 2 ** ((2**5 - 2 - get_default_exp_bias(torch.float8_e5m2))) * (2 - 2 ** -(8 - 1 - 5)),
}


def get_fullscale(dtype, exp_bias=None):
    default_exp_bias = get_default_exp_bias(dtype)
    fullscale = MAX_RANGE[dtype]
    exp_bias = default_exp_bias if exp_bias is None else exp_bias
    fullscale = fullscale * (2 ** (default_exp_bias - exp_bias))
    return fullscale


def get_fullscales_by_expbias_set(dtype, expbias_set):
    return [get_fullscale(dtype, exp_bias=eb) for eb in expbias_set]


def get_fp8_hw_alligned_scales(dtype, device):
    exp_bias_set = EXP_BIAS_SETS.get((device, dtype), None)
    return (
        None
        if exp_bias_set is None
        else [x / MAX_RANGE[dtype] for x in get_fullscales_by_expbias_set(dtype, exp_bias_set)]
    )


DEVICES_SCALE_FACTORS = {
    htexp.synDeviceType.synDeviceGaudi2: 4,
    htexp.synDeviceType.synDeviceGaudi3: 1,
}
FP8_143_SCALES = {
    device: get_fp8_hw_alligned_scales(torch.float8_e4m3fn, device) for device in DEVICES_SCALE_FACTORS.keys()
}
FP8_143_SCALES_TRAITS = {
    device: (
        min(FP8_143_SCALES[device]),
        max(FP8_143_SCALES[device]),
        DEVICES_SCALE_FACTORS[device],
    )
    for device in DEVICES_SCALE_FACTORS.keys()
}


def calc_maxabs_scale(xmaxabs, fullscale, backoff=1):
    scale = xmaxabs / (fullscale * backoff)
    return scale


def scale_to_pow2(scale):
    scale_pow2 = 2 ** torch.ceil(torch.log2(scale))
    return scale_pow2


# Considering range of hw aligned scales: 2^a, 2^a+1,..., 2^b (a<b)
# we want to choose scale s for maxabs m such that 2^a <= s=2^x <= 2^b (for integer a<=x<=b)
# and also 2^(x-1) < m <= 2^x
# if m>=2^b then s=2^b, therefore min(_, 2^b)
# if m<=2^a then s=2^a, therefore max(_, 2^a) --> 2^a <= min(max(_,2^a),2^b) <=2^b
# if s^a<m<2^b then m as a positive number can be written as m=2^y (y=log2(m))
# if y is integer then y=ciel(y) we choose x=y so s=2^x=2^y=2^ciel(y)=2^ciel(log2(m))
# else we choose x=ciel(y) and a<=x-1<y<x<=b and s=2^x=2^ciel(y)=2^ciel(log2(m))
# for Gaudi2 the range is 16^-2..16^1 so we change 2 with 16 and remember that:
# 16 = 2^4, log16(m)=log2(m)/log2(16)=log2(m)/4, and we get:
# we choose s=16^ciel(log16(m))=2^4^ciel(log2(m)/4)=2^(4*ciel(log2(m)/4))=2^(ciel(log2(m)/4)*4)
def scale_to_pow2_hw(scale, device_type):
    scale_pow2 = scale_to_pow2(scale)
    min_scale, max_scale, scale_factor = FP8_143_SCALES_TRAITS[device_type]
    scale_pow2_hw = torch.minimum(
        torch.maximum(
            2 ** (torch.ceil(torch.log2(scale_pow2) / scale_factor) * scale_factor),
            torch.tensor(min_scale, dtype=scale.dtype, device=scale.device),
        ),
        torch.tensor(max_scale, dtype=scale.dtype, device=scale.device),
    )
    return scale_pow2_hw


def mmse_scale_multi(x, ref_scale, scales, lp_dtype, hp_dtype):
    # TODO: SW-176672 move weights to hpu before the scale calculations
    x = x.to("hpu")
    Nch = x.shape[-1]
    opt_err = torch.ones(Nch, dtype=hp_dtype, device=x.device) * torch.inf
    opt_scale = torch.ones(Nch, dtype=hp_dtype, device=x.device) * -1
    sum_axis = list(range(x.ndim - 1))
    rs = ref_scale.unsqueeze(dim=1).transpose(0, 1)
    for s in scales:
        sv = torch.ones(Nch, dtype=hp_dtype, device=x.device) * s
        xscales = rs * sv
        y = scale_fcn(x, xscales)
        y = cast_to_fp8_fcn(y, lp_dtype)
        htcore.mark_step()  # we are measuring the error so we want to avoid fusion of the converts
        y = cast_fcn(y, hp_dtype)
        y = descale_fcn(y, xscales)
        err = torch.sum((x - y) ** 2, dim=sum_axis)
        opt_scale = torch.where(err < opt_err, sv, opt_scale)
        opt_err = torch.where(err < opt_err, err, opt_err)
        htcore.mark_step()
    return opt_scale * ref_scale


def mmse_scale(x, scales, lp_dtype, hp_dtype):
    # TODO: SW-176672 move weights to hpu before the scale calculations
    x = x.to("hpu")
    opt_err = torch.ones(1, dtype=hp_dtype, device=x.device) * torch.inf
    opt_scale = -1
    for s in scales:
        y = scale_fcn(x, s)
        y = cast_to_fp8_fcn(y, lp_dtype)
        htcore.mark_step()  # we are measuring the error so we want to avoid fusion of the converts
        y = cast_fcn(y, hp_dtype)
        y = descale_fcn(y, s)
        err = torch.norm(x - y)
        opt_scale = torch.where(err <= opt_err, s, opt_scale)
        opt_err = torch.where(err <= opt_err, err, opt_err)
        htcore.mark_step()
    return opt_scale


def manipulate_scales(scales, func):
    new_inputs = [func(input) for input in scales.inputs]
    new_outputs = [func(output) for output in scales.outputs]
    new_weights = {}
    if "weight" in scales.params.keys():
        if isinstance(scales.params["weight"], (torch.Tensor, float)):
            new_weights = {"weight": func(scales.params["weight"])}
        elif isinstance(scales.params["weight"], dict):
            new_weights_dict = {}
            for key, wt in scales.params["weight"].items():
                new_weights_dict[key] = func(wt)
            new_weights = {"weight": new_weights_dict}
    new_scales = ModuleConfig(new_inputs, new_outputs, new_weights)
    return new_scales


def invert_scales(scales):
    def invert(x):
        if isinstance(x, (list, tuple)):
            return [1 / x_i for x_i in x]
        return 1 / x

    return manipulate_scales(scales, invert)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch utilities."""


import enum
from typing import Callable, Dict, List, Optional, Tuple, Union

import psutil
import torch
from typing_extensions import TypeAlias

from neural_compressor.common.utils import (
    Mode,
    ProcessorType,
    Statistics,
    cpu_info,
    detect_processor_type_based_on_hw,
    logger,
)

OP_NAME_AND_TYPE_TUPLE_TYPE: TypeAlias = Tuple[str, Union[torch.nn.Module, Callable]]

# Dictionary to store a mapping between algorithm names and corresponding algo implementation(function)
algos_mapping: Dict[str, Callable] = {}


# All constants for torch
WHITE_MODULE_LIST = [torch.nn.Linear, torch.nn.Conv1d, torch.nn.Conv2d, torch.nn.Conv3d]

HPU_SAFE_WEIGHTS_NAME = "hpu_model.safetensors"
WEIGHT_NAME = "quantized_weight.pt"
HPU_WEIGHT_NAME = "quantized_hpu_weight.pt"
QCONFIG_NAME = "qconfig.json"


def register_algo(name):
    """Decorator function to register algorithms in the algos_mapping dictionary.

    Usage example:
        @register_algo(name=example_algo)
        def example_algo(model: torch.nn.Module, quant_config: RTNConfig) -> torch.nn.Module:
            ...

    Args:
        name (str): The name under which the algorithm function will be registered.

    Returns:
        decorator: The decorator function to be used with algorithm functions.
    """

    def decorator(algo_func):
        algos_mapping[name] = algo_func
        return algo_func

    return decorator


def fetch_module(model, op_name):
    """Get module with a given op name.

    Args:
        model (object): the input model.
        op_name (str): name of op.

    Returns:
        module (object).
    """
    module = model
    name_list = op_name.split(".")
    for name in name_list:
        if hasattr(module, name):
            module = getattr(module, name)
        else:
            logger.warning(f"The {op_name} is not present in the model.")
            return None
    return module


def set_module(model, op_name, new_module):
    """Set module with a given op name.

    Args:
        model (object): the input model.
        op_name (str): name of op.
        new_module (object): the input model.

    Returns:
        module (object).
    """
    name_list = op_name.split(".")
    if len(name_list) == 1:
        setattr(model, name_list[-1], new_module)
        return
    else:
        second_last_module = fetch_module(model, ".".join(name_list[:-1]))
        if second_last_module is None:
            logger.warning(f"Setting skipped as the {op_name} is not present in the model.")
            return None
        else:
            setattr(second_last_module, name_list[-1], new_module)


get_attr = fetch_module
set_attr = set_module


def get_model_info(model: torch.nn.Module, white_module_list: List[Callable]) -> List[Tuple[str, str]]:
    """Get model info according to white_module_list."""
    module_dict = dict(model.named_modules())
    filter_result = []
    filter_result_set = set()
    for op_name, module in module_dict.items():
        if isinstance(module, tuple(white_module_list)):
            pair = (op_name, type(module).__name__)
            if pair not in filter_result_set:
                filter_result_set.add(pair)
                filter_result.append(pair)
    logger.debug(f"Get model info: {filter_result}")
    return filter_result


def get_double_quant_config_dict(double_quant_type="BNB_NF4"):
    """Query config dict of double_quant according to double_quant_type.

    Args:
        double_quant_type (str, optional): double_quant type. Defaults to "BNB_NF4".
    """
    from neural_compressor.torch.utils.constants import DOUBLE_QUANT_CONFIGS

    assert double_quant_type in DOUBLE_QUANT_CONFIGS, "Supported double quant configs: {}".format(
        list(DOUBLE_QUANT_CONFIGS.keys())
    )
    return DOUBLE_QUANT_CONFIGS[double_quant_type]


def get_quantizer(model, quantizer_cls, quant_config=None, *args, **kwargs):
    """Get the quantizer.

    Initialize a quantizer or get `quantizer` attribute from model.

    Args:
        model (torch.nn.Module): pytorch model.
        quantizer_cls (Quantizer): quantizer class of a specific algorithm.
        quant_config (dict, optional): Specifies how to apply the algorithm on the given model.
            Defaults to None.

    Returns:
        quantizer object.
    """
    if not hasattr(model, "quantizer"):
        quantizer = quantizer_cls(quant_config=quant_config, *args, **kwargs)
        return quantizer
    else:
        return model.quantizer


def postprocess_model(model, mode, quantizer):
    """Process `quantizer` attribute of model according to current phase.

    In `prepare` phase, the `quantizer` is set as an attribute of the model
    to avoid redundant initialization during `convert` phase.

    In 'convert' or 'quantize' phase, the unused `quantizer` attribute is removed.

    Args:
        model (torch.nn.Module): pytorch model.
        mode (Mode): The mode of current phase, including 'prepare', 'convert' and 'quantize'.
        quantizer (Quantizer): quantizer object.
    """
    if mode == Mode.PREPARE:
        model.quantizer = quantizer
    elif mode == Mode.CONVERT or mode == Mode.QUANTIZE:
        if getattr(model, "quantizer", False):
            del model.quantizer


def dump_model_op_stats(mode, tune_cfg):
    """Dump quantizable ops stats of model to user.

    Args:
        mode (object): quantization mode.
        tune_cfg (dict): quantization config
    """
    if mode == Mode.PREPARE:
        return
    res = {}
    # collect all dtype info and build empty results with existing op_type
    dtype_set = set()
    for op, config in tune_cfg.items():
        op_type = op[1]
        config = config.to_dict()
        if not config["dtype"] == "fp32":
            num_bits = config["bits"]
            group_size = config["group_size"]
            dtype_str = "A32W{}G{}".format(num_bits, group_size)
            dtype_set.add(dtype_str)
    dtype_set.add("FP32")
    dtype_list = list(dtype_set)
    dtype_list.sort()

    for op, config in tune_cfg.items():
        config = config.to_dict()
        op_type = op[1]
        if op_type not in res.keys():
            res[op_type] = {dtype: 0 for dtype in dtype_list}

    # fill in results with op_type and dtype
    for op, config in tune_cfg.items():
        config = config.to_dict()
        if config["dtype"] == "fp32":
            res[op_type]["FP32"] += 1
        else:
            num_bits = config["bits"]
            group_size = config["group_size"]
            dtype_str = "A32W{}G{}".format(num_bits, group_size)
            res[op_type][dtype_str] += 1

    # update stats format for dump.
    field_names = ["Op Type", "Total"]
    field_names.extend(dtype_list)
    output_data = []
    for op_type in res.keys():
        field_results = [op_type, sum(res[op_type].values())]
        field_results.extend([res[op_type][dtype] for dtype in dtype_list])
        output_data.append(field_results)

    Statistics(output_data, header="Mixed Precision Statistics", field_names=field_names).print_stat()


def get_model_device(model: torch.nn.Module):
    """Get the device.

    Args:
        model (torch.nn.Module): the input model.

    Returns:
        device (str): a string.
    """
    for n, p in model.named_parameters():
        return p.data.device.type  # p.data.device == device(type='cpu')


def get_processor_type_from_user_config(user_processor_type: Optional[Union[str, ProcessorType]] = None):
    """Get the processor type.

    Get the processor type based on the user configuration or automatically detect it based on the hardware.

    Args:
        user_processor_type (Optional[Union[str, ProcessorType]]): The user-specified processor type. Defaults to None.

    Returns:
        ProcessorType: The detected or user-specified processor type.

    Raises:
        AssertionError: If the user-specified processor type is not supported.
        NotImplementedError: If the processor type is not recognized.
    """
    if user_processor_type is None:
        processor_type = detect_processor_type_based_on_hw()
    elif isinstance(user_processor_type, ProcessorType):
        processor_type = user_processor_type
    elif isinstance(user_processor_type, str):
        user_processor_type = user_processor_type.lower().capitalize()
        assert user_processor_type in ProcessorType.__members__, f"Unsupported processor type: {user_processor_type}"
        processor_type = ProcessorType(user_processor_type)
    else:
        raise NotImplementedError(f"Unsupported processor type: {user_processor_type}")
    return processor_type


def dowload_hf_model(repo_id, cache_dir=None, repo_type=None, revision=None):
    """Download hugging face model from hf hub."""
    import os

    from huggingface_hub.constants import DEFAULT_REVISION, HUGGINGFACE_HUB_CACHE
    from huggingface_hub.file_download import REGEX_COMMIT_HASH, repo_folder_name
    from huggingface_hub.utils import EntryNotFoundError

    if cache_dir is None:
        cache_dir = HUGGINGFACE_HUB_CACHE
    if revision is None:
        revision = DEFAULT_REVISION
    if repo_type is None:
        repo_type = "model"
    storage_folder = os.path.join(cache_dir, repo_folder_name(repo_id=repo_id, repo_type=repo_type))
    commit_hash = None
    if REGEX_COMMIT_HASH.match(revision):
        commit_hash = revision
    else:
        ref_path = os.path.join(storage_folder, "refs", revision)
        if os.path.exists(ref_path):
            with open(ref_path) as f:
                commit_hash = f.read()
    if storage_folder and commit_hash:
        pointer_path = os.path.join(storage_folder, "snapshots", commit_hash)
        if os.path.isdir(pointer_path):
            return pointer_path
    else:  # pragma: no cover
        from huggingface_hub import snapshot_download

        file_path = snapshot_download(repo_id)
        return file_path


def load_empty_model(pretrained_model_name_or_path, cls=None, **kwargs):
    """Load a empty model."""
    import os

    from accelerate import init_empty_weights
    from transformers import AutoConfig, AutoModelForCausalLM
    from transformers.models.auto.auto_factory import _BaseAutoModelClass

    cls = AutoModelForCausalLM if cls is None else cls
    is_local = os.path.isdir(pretrained_model_name_or_path)
    if is_local:  # pragma: no cover
        path = pretrained_model_name_or_path
    else:
        path = dowload_hf_model(pretrained_model_name_or_path)
    if cls.__base__ == _BaseAutoModelClass:
        config = AutoConfig.from_pretrained(path, **kwargs)
        with init_empty_weights():
            model = cls.from_config(config)
    else:  # pragma: no cover
        config = cls.config_class.from_pretrained(path, **kwargs)
        with init_empty_weights():
            model = cls(config)
    model.tie_weights()
    model.eval()
    model.path = pretrained_model_name_or_path
    return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from enum import Enum, Flag, auto
from json.decoder import JSONDecodeError
from typing import Any, Mapping

import habana_frameworks.torch.utils.experimental as htexp
import torch

from ..utils.logger import logger

local_rank = int(os.getenv("LOCAL_RANK", "-1"))
world_size = int(os.getenv("WORLD_SIZE", "-1"))


class QuantMode(Enum):
    NONE = 0
    QUANTIZE = 1
    MEASURE = 2
    SHAPE = 3


class MeasureExclude(Flag):
    NONE = auto()
    INPUT = auto()
    OUTPUT = auto()
    PARAMS = auto()
    ALL = auto()


class ScaleMethod(Enum):
    MAX = 1
    UNIT_SCALE = 2
    MAXABS_HW = 3
    MAXABS_POW2 = 4
    SMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2 = 5
    WEAKSMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2 = 6
    ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2 = 7
    ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2 = 8
    ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2 = 9
    ACT_MAXABS_POW2_WEIGHTS_PCS_OPT_POW2 = 10
    SMOOTHQUANT_OPT = 11
    MAXABS_HW_OPT_WEIGHT = 12
    MAXABS_POW2_OPT_WEIGHT = 13


def get_hqt_config(mod) -> Fp8cfg:
    return mod.__hqt_config__


def set_hqt_config(mod, config):
    mod.__hqt_config__ = config


@dataclass
class Fp8cfg:
    cfg: Mapping[str, Any]

    def parse(custom_config: Mapping[str, str]) -> Fp8cfg:
        measured_global_config = {
            "dump_stats_path": "stats",
            "fp8_config": torch.float8_e4m3fn,  # The parameters of the chosen Quantization methed
            "hp_dtype": torch.bfloat16,  # The parameters of the chosen Quantization methed
            "blocklist": {
                "names": [],
                "types": (),
            },  # types and names to not be quantized
            "allowlist": {
                "names": [],
                "types": ("torch.nn.Linear", "torch.nn.Conv2d", "BMM"),
            },  # types and names to be quantized. Allowlist by names is not yet implemented
            "mode": QuantMode.QUANTIZE,  # Quantize or Measure
            "scale_method": ScaleMethod.UNIT_SCALE,  # Method to quantize with
            "scale_params": {},  # scaling parameters that are different then the default ones
            "observer": "maxabs",  # Supported ['shape', 'maxabs', 'maxabs_per_channel', 'save']
            "mod_dict": {},
            "ignore_modules_wo_measures": False,  # Determines whether to fail quantization on modules without existing measures or not to quantize them
            "local_rank": local_rank if local_rank >= 0 else None,
            "global_rank": None,
            "world_size": world_size if world_size >= 0 else None,
            "seperate_measure_files": True,  # Determines whether to expect one or several measure files when using more than one gaudi
            "device_type": htexp._get_device_type(),  # Determines device type: Gaudi2, Gaudi3...
            "measure_exclude": MeasureExclude.OUTPUT,
        }
        # assert measured_global_config['allowlist']['names'] == [''], "Allowlist names not yet implemented"

        # go over all user-defined keys from json, handle various cases
        for keys in custom_config:
            if keys == "mode":
                if custom_config[keys] == "NONE":
                    custom_config[keys] = QuantMode.NONE
                elif custom_config[keys] == "QUANTIZE":
                    custom_config[keys] = QuantMode.QUANTIZE
                elif custom_config[keys] == "MEASURE":
                    custom_config[keys] = QuantMode.MEASURE
                elif custom_config[keys] == "SHAPE":
                    custom_config[keys] = QuantMode.SHAPE
                else:
                    raise ValueError("invalid mode in custom config. Enter Quantize or Measure")

            if keys == "measure_exclude":
                if custom_config[keys] == "NONE":
                    custom_config[keys] = MeasureExclude.NONE
                elif custom_config[keys] == "OUTPUT":
                    custom_config[keys] = MeasureExclude.OUTPUT
                elif custom_config[keys] == "INPUT":
                    custom_config[keys] = MeasureExclude.INPUT
                elif custom_config[keys] == "ALL":
                    custom_config[keys] = MeasureExclude.ALL
                else:
                    raise ValueError("invalid measure exclude value in custom config. Enter OUTPUT or NONE")

            if keys == "fp8_config":
                if custom_config[keys].lower() == "e4m3":
                    custom_config[keys] = torch.float8_e4m3fn

                elif custom_config[keys].lower() == "e5m2":
                    custom_config[keys] = torch.float8_e5m2
                else:
                    raise ValueError("invalid fp8_config in custom config. Enter E4M3 or E5M2")

            if keys == "hp_dtype":
                if custom_config[keys].lower() == "bf16":
                    custom_config[keys] = torch.bfloat16
                elif custom_config[keys].lower() == "fp16":
                    custom_config[keys] = torch.float16
                elif custom_config[keys].lower() == "fp32":
                    custom_config[keys] = torch.float32
                else:
                    raise ValueError("invalid hp_dtype in custom config. Enter bf16, fp16 or fp32")

            if keys == "scale_method":
                if custom_config[keys].lower() == "unit_scale":
                    custom_config[keys] = ScaleMethod.UNIT_SCALE
                elif custom_config[keys].lower() == "max":
                    custom_config[keys] = ScaleMethod.MAX
                elif custom_config[keys].lower() == "maxabs_hw":
                    custom_config[keys] = ScaleMethod.MAXABS_HW
                elif custom_config[keys].lower() == "maxabs_pow2":
                    custom_config[keys] = ScaleMethod.MAXABS_POW2
                elif custom_config[keys].lower() == "maxabs_hw_opt_weight":
                    custom_config[keys] = ScaleMethod.MAXABS_HW_OPT_WEIGHT
                elif custom_config[keys].lower() == "maxabs_pow2_opt_weight":
                    custom_config[keys] = ScaleMethod.MAXABS_POW2_OPT_WEIGHT
                elif custom_config[keys].lower() == "smoothquant_weights_output_channel_maxabs_pow2":
                    custom_config[keys] = ScaleMethod.SMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2
                elif custom_config[keys].lower() == "weaksmoothquant_weights_output_channel_maxabs_pow2":
                    custom_config[keys] = ScaleMethod.WEAKSMOOTHQUANT_WEIGHTS_OUTPUT_CHANNEL_MAXABS_POW2
                elif custom_config[keys].lower() == "act_maxabs_hw_weights_pcs_maxabs_pow2":
                    custom_config[keys] = ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_MAXABS_POW2
                elif custom_config[keys].lower() == "act_maxabs_hw_weights_pcs_opt_pow2":
                    custom_config[keys] = ScaleMethod.ACT_MAXABS_HW_WEIGHTS_PCS_OPT_POW2
                elif custom_config[keys].lower() == "act_maxabs_pow2_weights_pcs_maxabs_pow2":
                    custom_config[keys] = ScaleMethod.ACT_MAXABS_POW2_WEIGHTS_PCS_MAXABS_POW2
                elif custom_config[keys].lower() == "act_maxabs_pow2_weights_pcs_opt_pow2":
                    custom_config[keys] = ScaleMethod.ACT_MAXABS_POW2_WEIGHTS_PCS_OPT_POW2
                elif custom_config[keys].lower() == "smoothquant_opt":
                    custom_config[keys] = ScaleMethod.SMOOTHQUANT_OPT
                else:
                    raise ValueError(
                        f'Invalid fp8_config in custom config ({custom_config[keys]}). should be in ["max", "unit_scale", "maxabs_hw", "maxabs_pow2", "maxabs_per_channel_pow2", "smoothquant_opt"]'
                    )

            if keys == "ignore_modules_wo_measures":
                custom_config[keys] = custom_config[keys].lower() == "true"

            # TODO [SW-175936] - remove checking for old key names whitelist and blacklist.
            if isinstance(custom_config[keys], dict):
                for keys_2 in custom_config[keys]:
                    if keys == "whitelist":
                        measured_global_config["allowlist"][keys_2] = custom_config[keys][keys_2]
                    elif keys == "blacklist":
                        measured_global_config["blocklist"][keys_2] = custom_config[keys][keys_2]
                    else:
                        measured_global_config[keys][keys_2] = custom_config[keys][keys_2]
            else:
                if keys == "whitelist":
                    measured_global_config["allowlist"] = custom_config[keys]
                elif keys == "blacklist":
                    measured_global_config["blocklist"] = custom_config[keys]
                else:
                    measured_global_config[keys] = custom_config[keys]

        # If seperate_measure_files is True (default value), then it is assumed that there are multiple distinct measure and scale files
        # and they are stored in / loaded from paths with the correct index as a suffix. Else, only one is searched for.
        measured_global_config["local_rank"] = (
            local_rank if local_rank >= 0 and custom_config.get("seperate_measure_files", True) else None
        )

        base_name = measured_global_config["dump_stats_path"].split("/")[-1]
        folder_name = measured_global_config["dump_stats_path"][: -(len(base_name))]
        measured_global_config["dump_stats_base_path"] = folder_name
        os.makedirs(folder_name, exist_ok=True)
        worker_st = (
            ""
            if measured_global_config["local_rank"] is None
            else "_" + str(measured_global_config["local_rank"]) + "_" + str(measured_global_config["world_size"])
        )
        measured_global_config["shape_file"] = measured_global_config["dump_stats_path"] + "_hooks_shape" + worker_st
        measured_global_config["scale_file"] = (
            measured_global_config["dump_stats_path"]
            + "_hooks_"
            + measured_global_config["observer"]
            + "_"
            + measured_global_config["scale_method"].name
            + worker_st
        )
        if (measured_global_config["mode"] == QuantMode.MEASURE) or (
            measured_global_config["mode"] == QuantMode.QUANTIZE
        ):
            measured_global_config["measure_file"] = (
                measured_global_config["dump_stats_path"] + "_hooks_" + measured_global_config["observer"] + worker_st
            )
        # measured_global_config['dump_stats_path'] += '_hooks_.json'

        logger.debug("HQT Paths:")
        logger.debug("base_name='%s'", base_name)
        logger.debug("folder_name='%s'", folder_name)
        logger.debug(
            "measured_global_config['shape_file']='%s'",
            measured_global_config["shape_file"],
        )
        logger.debug(
            "measured_global_config['scale_file']='%s'",
            measured_global_config["scale_file"],
        )
        if "measure_file" in measured_global_config.keys():
            logger.debug(
                "measured_global_config['measure_file']='%s'",
                measured_global_config["measure_file"],
            )
        logger.debug(
            "measured_global_config['dump_stats_path']='%s'",
            measured_global_config["dump_stats_path"],
        )

        return Fp8cfg(cfg=measured_global_config)


def _read_config_from_file(config_path: str) -> Mapping[str, str]:
    logger.debug("QUANT PACKAGE: using %s config", config_path)

    module_directory = os.path.dirname(os.path.abspath(__file__))

    # if file in absolute path doesn't exist, try looking in cfg directory
    if not os.path.isfile(config_path):
        config_path = os.path.join(module_directory, "..", f"custom_config/{config_path}.json")
    try:
        logger.info("QUANT PACKAGE: Loading %s", config_path)
        with open(config_path) as config_json:
            config = json.load(config_json)
    except FileNotFoundError as e:
        raise Exception(f"Got exception: {e}. QUANT PACKAGE: Can't open {config_path}!")
    except JSONDecodeError as e:
        config_json.close()
        raise Exception(f"Got exception: {e}. QUANT PACKAGE: Can't load {config_path} json!")
    return config



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from ..common import *
from ..fp_utils import *


def linear_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    weight_scale = calc_maxabs_scale(
        torch.max(torch.abs(mod.weight.detach())).to(dtype=hp_dtype, device=device),
        fullscale,
        weight_backoff,
    )
    input_scale = scale_to_pow2_hw(input_scale, device_type=config["device_type"])
    weight_scale = scale_to_pow2_hw(weight_scale, device_type=config["device_type"])
    output_scale = input_scale * weight_scale
    return ModuleConfig((input_scale,), (output_scale,), {"weight": weight_scale})


def linear_act_maxabs_pts_weight_maxabs_pts_pow2_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    weight_scale = calc_maxabs_scale(
        torch.max(torch.abs(mod.weight.detach())).to(dtype=hp_dtype, device=device),
        fullscale,
        weight_backoff,
    )
    input_scale = scale_to_pow2(input_scale)
    weight_scale = scale_to_pow2(weight_scale)
    output_scale = input_scale * weight_scale
    return ModuleConfig((input_scale,), (output_scale,), {"weight": weight_scale})


def matmul_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = [
        calc_maxabs_scale(
            torch.tensor(x, dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
        for x in measurement.inputs
    ]
    input_scale = [scale_to_pow2_hw(x, device_type=config["device_type"]) for x in input_scale]
    output_scale = [input_scale[0] * input_scale[1]]
    return ModuleConfig(input_scale, output_scale, {})


def matmul_act_maxabs_pts_weight_maxabs_pts_pow2_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = [
        calc_maxabs_scale(
            torch.tensor(x, dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
        for x in measurement.inputs
    ]
    input_scale = [scale_to_pow2(x) for x in input_scale]
    output_scale = [input_scale[0] * input_scale[1]]
    return ModuleConfig(input_scale, output_scale, {})


def fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = [
        calc_maxabs_scale(
            torch.tensor(x, dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
        for x in measurement.inputs
    ]
    # add amax scale to input scales
    input_scale.append(
        calc_maxabs_scale(
            torch.tensor(measurement.outputs[1], dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
    )
    input_scale = [scale_to_pow2_hw(x, device_type=config["device_type"]) for x in input_scale]
    output_scale = calc_maxabs_scale(
        torch.tensor(measurement.outputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    output_scale = [scale_to_pow2_hw(output_scale, device_type=config["device_type"])]
    return ModuleConfig(input_scale, output_scale, {})


def fsdpa_act_maxabs_pts_weight_maxabs_pts_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = [
        calc_maxabs_scale(
            torch.tensor(x, dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
        for x in measurement.inputs
    ]
    # fsdpa is combined out of - BMM1(Q,K) -> Softmax -> BMM2(AMAX,V)
    # during measure we receive the amax value from the cguid and apply it during quant as input
    input_scale.append(
        calc_maxabs_scale(
            torch.tensor(measurement.outputs[1], dtype=hp_dtype, device=device).max(),
            fullscale,
            input_backoff,
        )
    )
    input_scale = [scale_to_pow2(x) for x in input_scale]
    output_scale = calc_maxabs_scale(
        torch.tensor(measurement.outputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    output_scale = [scale_to_pow2(output_scale)]
    return ModuleConfig(input_scale, output_scale, {})


def linear_act_maxabs_pts_weight_opt_pts_pow2_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    scales = params["weight_scales"]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    weight_scale = mmse_scale(mod.weight, scales, lp_dtype, hp_dtype)
    input_scale = scale_to_pow2(input_scale)
    weight_scale = scale_to_pow2(weight_scale)
    output_scale = input_scale * weight_scale
    return ModuleConfig((input_scale,), (output_scale,), {"weight": weight_scale})


def linear_act_maxabs_pts_weight_opt_pts_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    scales = params["weight_scales"]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    weight_scale = mmse_scale(mod.weight, scales, lp_dtype, hp_dtype)
    input_scale = scale_to_pow2_hw(input_scale, device_type=config["device_type"])
    weight_scale = scale_to_pow2_hw(weight_scale, device_type=config["device_type"])
    output_scale = input_scale * weight_scale
    return ModuleConfig((input_scale,), (output_scale,), {"weight": weight_scale})


def kv_cache_act_maxabs_pts_weight_maxabs_pts_pow2_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    # calc the scale per layer tensor
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale_list = [scale_to_pow2_hw(input_scale, device_type=config["device_type"])]
    output_scale = [input_scale_list[0]]  # output scale is same as the first input (current data) since range is same
    return ModuleConfig(input_scale_list, output_scale, {})


def kv_cache_act_maxabs_pts_pow2_weight_opt_pcs_pow2_scales(mod, measurement, params):
    # calc the scale per layer tensor
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale_list = [scale_to_pow2(input_scale)]
    output_scale = [input_scale_list[0]]  # output scale is same as the first input (current data) since range is same
    return ModuleConfig(input_scale_list, output_scale, {})


def softmax_input_unit_output_maxabs_pts_hw_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    output_scale = calc_maxabs_scale(
        torch.tensor(measurement.outputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    output_scale = [scale_to_pow2_hw(output_scale, device_type=config["device_type"])]
    return ModuleConfig((), output_scale, {})


def linear_act_maxabs_pts_pow2_hw_weights_maxabs_pcs_pow2_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale = scale_to_pow2_hw(input_scale, device_type=config["device_type"])
    weight_scale_in_ch = torch.ones([mod.weight.shape[1], 1], dtype=hp_dtype, device=device)

    weight_range_out_ch = torch.max(torch.abs(mod.weight), dim=1)[0].reshape([-1, 1])
    weight_maxabs_scale_out_ch = calc_maxabs_scale(weight_range_out_ch, fullscale, weight_backoff)
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)
    output_scale = weight_maxabs_scale_out_ch * input_scale
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {
            "weight": {
                0: weight_maxabs_scale_out_ch.flatten(),
                1: weight_scale_in_ch.flatten(),
            }
        },
    )


def linear_act_maxabs_pts_pow2_weights_maxabs_pcs_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale = scale_to_pow2(input_scale)
    weight_scale_in_ch = torch.ones([mod.weight.shape[1], 1], dtype=hp_dtype, device=device)

    weight_range_out_ch = torch.max(torch.abs(mod.weight), dim=1)[0].reshape([-1, 1])
    weight_maxabs_scale_out_ch = calc_maxabs_scale(weight_range_out_ch, fullscale, weight_backoff)
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)
    output_scale = weight_maxabs_scale_out_ch * input_scale
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {
            "weight": {
                0: weight_maxabs_scale_out_ch.flatten(),
                1: weight_scale_in_ch.flatten(),
            }
        },
    )


def linear_act_maxabs_pts_pow2_hw_weights_opt_pcs_pow2_scales(mod, measurement, params):
    config = get_hqt_config(mod).cfg
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    weight_scales = params["weight_scales"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale = scale_to_pow2_hw(input_scale, device_type=config["device_type"])
    weight_scale_in_ch = torch.ones([mod.weight.shape[1], 1], dtype=hp_dtype, device=device)

    weight_range_out_ch = torch.max(torch.abs(mod.weight), dim=1)[0].reshape([-1, 1])
    weight_maxabs_scale_out_ch = calc_maxabs_scale(weight_range_out_ch, fullscale, weight_backoff)
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)
    weight_opt_scale_out_ch = mmse_scale_multi(
        torch.transpose(mod.weight, 0, 1),
        weight_maxabs_scale_out_ch.squeeze(),
        weight_scales,
        lp_dtype,
        hp_dtype,
    ).unsqueeze(1)
    weight_maxabs_scale_out_ch = weight_opt_scale_out_ch
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)  # should be power of 2, just making sure
    output_scale = weight_maxabs_scale_out_ch * input_scale
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {
            "weight": {
                0: weight_maxabs_scale_out_ch.flatten(),
                1: weight_scale_in_ch.flatten(),
            }
        },
    )


def linear_act_maxabs_pts_pow2_weights_opt_pcs_pow2_scales(mod, measurement, params):
    device = torch.device("hpu")
    lp_dtype = params["lp_dtype"]
    hp_dtype = params["hp_dtype"]
    fullscale = MAX_RANGE[lp_dtype]
    input_backoff = params["input_backoff"]
    weight_backoff = params["weight_backoff"]
    weight_scales = params["weight_scales"]
    input_scale = calc_maxabs_scale(
        torch.tensor(measurement.inputs[0], dtype=hp_dtype, device=device).max(),
        fullscale,
        input_backoff,
    )
    input_scale = scale_to_pow2(input_scale)
    weight_scale_in_ch = torch.ones([mod.weight.shape[1], 1], dtype=hp_dtype, device=device)

    weight_range_out_ch = torch.max(torch.abs(mod.weight), dim=1)[0].reshape([-1, 1])
    weight_maxabs_scale_out_ch = calc_maxabs_scale(weight_range_out_ch, fullscale, weight_backoff)
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)
    weight_opt_scale_out_ch = mmse_scale_multi(
        torch.transpose(mod.weight, 0, 1),
        weight_maxabs_scale_out_ch.squeeze(),
        weight_scales,
        lp_dtype,
        hp_dtype,
    ).unsqueeze(1)
    weight_maxabs_scale_out_ch = weight_opt_scale_out_ch
    weight_maxabs_scale_out_ch = scale_to_pow2(weight_maxabs_scale_out_ch)  # should be power of 2, just making sure
    output_scale = weight_maxabs_scale_out_ch * input_scale
    return ModuleConfig(
        (input_scale.flatten(),),
        (output_scale.flatten(),),
        {
            "weight": {
                0: weight_maxabs_scale_out_ch.flatten(),
                1: weight_scale_in_ch.flatten(),
            }
        },
    )



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn

from .quant_config import QuantMode, get_hqt_config

try:  # backwards compatibility for 1.16
    from habana_frameworks.torch.hpex.kernels import fp8_fused_sdpa
except ImportError:
    pass


class BMM(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, y):
        return torch.bmm(x, y)


class Matmul(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, *args, **kwargs):
        return torch.matmul(*args, **kwargs)


class Identity(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x.clone()


class Softmax(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, dim=None):
        return torch.softmax(x, dim)


def matmul_fp8(
    input,
    other,
    out=None,
    out_dtype=torch.bfloat16,
    scale_input_inv=None,
    scale_other_inv=None,
):
    res = torch.ops.hpu.fp8_gemm_v2(
        input,
        False,
        other,
        False,
        out,
        out_dtype,
        scale_input_inv,
        scale_other_inv,
        None,
        False,
    )
    return res


def measure_input(input, observer):
    for i in range(len(observer)):
        observer[i].measure(input[i])


def measure_output(output, observer):
    if observer:
        for i in range(len(observer)):
            observer[i].measure(output[i])


def conv2d_fp8(
    input,
    other,
    bias,
    stride,
    padding,
    dilation,
    groups,
    out_dtype=torch.bfloat16,
    scale_input_inv=None,
    scale_other_inv=None,
):
    return torch.ops.hpu.conv2d_fp8(
        input=input,
        weight=other,
        bias=bias,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        out_dtype=out_dtype,
        scale_input=scale_input_inv,
        scale_weight=scale_other_inv,
    )


def set_attrs_from_orig_model(cls_instance, mod, mod_extra_config, *func_names):
    cls_instance.__dict__.update(mod.__dict__)
    config = get_hqt_config(cls_instance)
    cls_instance.extra_repr_org = mod.extra_repr
    cls_instance.class_name_org = mod.__class__.__name__
    cls_instance._mod_extra_config = mod_extra_config
    cls_instance.quantization_mode = config.cfg["mode"]
    # store original module in order to invoke its functions during measurements.
    # this may be omitted of torch remove the related validation from dynamo. see SW-187731.
    cls_instance.__dict__["orig_mod"] = mod
    cls_instance.forward_orig = mod.forward
    if func_names is not None:
        for func in func_names:
            setattr(cls_instance, func, getattr(mod, func))


def get_current_repr(cls_instance, *member_names):
    curr_repr = ""
    if cls_instance.quantization_mode == QuantMode.QUANTIZE:
        first_name = True
        for name in member_names:
            if not first_name:
                curr_repr += ", "
            curr_repr += f"{name} dtype={getattr(cls_instance, name).dtype}"
            first_name = False
    return curr_repr


def extra_representation(org_repr, org_name, curr_repr):
    repr = f"original={org_name}," + (" " + org_repr + "," if org_repr != "" else "")
    return f"{repr} {curr_repr}"


def _raise_lora_layer_error(layer_class):
    raise RuntimeError(
        f"{layer_class} quantization is not supported in case of lora_layer member is not None."
        f" Can add {layer_class} to 'blocklist' field in quantization config file"
    )


class PatchedMatmul(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input_0 = self._mod_extra_config.inputs[0]
            self.quant_input_1 = self._mod_extra_config.inputs[1]
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            self.scale_other = nn.Parameter(mod_extra_config.scale.inputs[1])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input, other):
        qinput = self.quant_input_0(input)
        qother = self.quant_input_1(other)
        output = matmul_fp8(
            qinput,
            qother,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_other,
        )
        return output

    def forward_measure(self, input, other):
        measure_input((input, other), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, other)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_other"),
        )


class PatchedLinear(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            # When offloading weights to disk using device_map, the module forward is overridden.
            # __dict__.update call again overrides the PatchedLinear forward with the forward that device_map planted.
            # So need to set PatchedLinear forawrd to be the right forward.
            self.forward = self.forward_quant
            self.quant_input = self._mod_extra_config.inputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
            elif isinstance(mod_extra_config.scale.params["weight"], dict):
                # PCQ weight is calculated with actual weight [0] and ones [1]
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"][0])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward_quant(self, input):
        qinput = self.quant_input(input)
        y = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        output = y + self.bias if (self.bias is not None) else y
        return output

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedLinearAllReduce(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        self.scoped_version = mod.__class__.__name__ == "ScopedLinearAllReduce"
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
            elif isinstance(mod_extra_config.scale.params["weight"], dict):
                # PCQ weight is calculated with actual weight [0] and ones [1]
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"][0])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input):
        # pre_all_reduce
        qinput = self.quant_input(input)
        output = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        dqoutput = self.quant_output(output)
        if not self.scoped_version:
            self.all_reduce(dqoutput)
            dqoutput = self.post_all_reduce(dqoutput)
        return dqoutput

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(input, self.weight.transpose(-1, -2))
        measure_output((output,), self._mod_extra_config.outputs)
        # in scoped version all reduce is being called outside of the layer
        if not self.scoped_version:
            self.all_reduce(output)
            output = self.post_all_reduce(output)
        return output

    def all_reduce(self, input):
        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(input, group=self.mp_group)

    def post_all_reduce(self, input):
        output = input + self.bias if (self.bias is not None) else input
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedRowParallelLinear(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config, "resolve_input")
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
            elif isinstance(mod_extra_config.scale.params["weight"], dict):
                # PCQ weight is calculated with actual weight [0] and ones [1]
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"][0])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input):
        resolved_input = self.resolve_input(input)
        qinput = self.quant_input(resolved_input)
        output = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        dqoutput = self.quant_output(output)
        if self.reduce_results:
            dqoutput = self.collective_func(dqoutput)
        return self.post_all_reduce(dqoutput)

    def forward_measure(self, input):
        resolved_input = self.resolve_input(input)
        measure_input((resolved_input,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(resolved_input, self.weight.transpose(-1, -2))
        measure_output((output,), self._mod_extra_config.outputs)
        if self.reduce_results:
            output = self.collective_func(output)
        return self.post_all_reduce(output)

    def post_all_reduce(self, output):
        assert (
            self.reduce_results or (not self.bias) or self.skip_bias_add
        ), "When not reduce the results, adding bias to the results can lead to incorrect results"
        if not self.skip_bias_add:
            output = output + self.bias if self.bias is not None else output
            output_bias = None
        else:
            output_bias = self.bias
        return output, output_bias

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedColumnParallelLinear(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
            elif isinstance(mod_extra_config.scale.params["weight"], dict):
                # PCQ weight is calculated with actual weight [0] and ones [1]
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"][0])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input):
        qinput = self.quant_input(input)
        output = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        dqoutput = self.quant_output(output)
        if self.gather_output:
            dqoutput = self.orig_mod.collective_func(dqoutput)
        return self.post_all_reduce(dqoutput)

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(input, self.weight.transpose(-1, -2))
        measure_output((output,), self._mod_extra_config.outputs)
        if self.gather_output:
            output = self.orig_mod.collective_func(output)
        return self.post_all_reduce(output)

    def post_all_reduce(self, output):
        if not self.skip_bias_add:
            output = output + self.bias if self.bias is not None else output
            output_bias = None
        else:
            output_bias = self.bias
        return output, output_bias

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedLmHeadLinearAllreduce(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            if isinstance(mod_extra_config.scale.params["weight"], (torch.Tensor, float)):
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
            elif isinstance(mod_extra_config.scale.params["weight"], dict):
                # PCQ weight is calculated with actual weight [0] and ones [1]
                self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"][0])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input):
        assert (
            input.shape[-1] % self.world_size == 0
        ), "Please ensure that self.world_size is divisible by input.shape[-1]"
        input_shard = input.shape[-1] // self.world_size
        splittedInput = input[:, :, self.rank * input_shard : (self.rank + 1) * input_shard]
        qinput = self.quant_input(splittedInput)
        output = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        dqoutput = self.quant_output(output)

        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(dqoutput, group=self.mp_group)
        if self.bias is not None:
            dqoutput += self.bias
        return dqoutput

    def forward_measure(self, input):
        assert (
            input.shape[-1] % self.world_size == 0
        ), "Please ensure that self.world_size is divisible by input.shape[-1]"
        input_shard = input.shape[-1] // self.world_size
        splittedInput = input[:, :, self.rank * input_shard : (self.rank + 1) * input_shard]
        measure_input((splittedInput,), observer=self._mod_extra_config.inputs)
        output = torch.matmul(splittedInput, self.weight.t())
        measure_output((output,), self._mod_extra_config.outputs)

        if self.mp_group is not None:
            from deepspeed import comm as dist

            dist.inference_all_reduce(output, group=self.mp_group)
        if self.bias is not None:
            output += self.bias
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedKVCache(nn.Module):
    # Module to patch KVCache module from llama model
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config, "forward", "get_shape")
        self.org_allocate = mod.allocate
        self.org_update = mod.update
        if self.quantization_mode == QuantMode.QUANTIZE:
            mod.update = self.update
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.update = self.update_measure
            mod.update = self.update_measure

    # overwrite allocate function of original module to force allocation in fp8
    def allocate(self, inp_seq_len, dtype, device, shape):
        dtype = torch.float8_e4m3fn if (self.quantization_mode == QuantMode.QUANTIZE) else dtype
        return self.org_allocate(inp_seq_len, dtype, device, shape)

    # overwrite update function of original module to force quant and dequant of cache input and output
    def update(self, prev, cur, dim, idx, inp_seq_len):
        qinput = self.quant_input(cur)
        output = self.org_update(prev, qinput, dim, idx, inp_seq_len)
        if output.dtype == torch.float8_e4m3fn:
            return self.quant_output(output)
        else:
            return output

    # overwrite update function of original module to force quant and dequant of cache input and output
    def update_measure(self, prev, cur, dim, idx, inp_seq_len):
        measure_input((cur,), self._mod_extra_config.inputs)
        output = self.org_update(prev, cur, dim, idx, inp_seq_len)
        measure_output((output,), self._mod_extra_config.outputs)
        return output


class PatchedVLLMKVCache(nn.Module):
    # Module to patch VLLMKVCache module from llama model
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.quant_output = self._mod_extra_config.outputs[0]
            self.orig_fetch_from_cache = mod.fetch_from_cache
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.fetch_from_cache = mod.fetch_from_cache
            self.forward = self.forward_measure

    def forward(self, input, cache, block_indices, block_offset):
        qinput = self.quant_input(input)
        output_cache = self.forward_orig(qinput, cache, block_indices, block_offset)
        return self.quant_output(output_cache)

    def forward_measure(self, input, cache, block_indices, block_offset):
        measure_input((input), self._mod_extra_config.inputs)
        output_cache = self.forward_orig(input, cache, block_indices, block_offset)
        measure_output((output_cache), self._mod_extra_config.outputs)
        return output_cache

    def fetch_from_cache(self, cache, blocks, permutations):
        quant_cache = self.quant_input(cache)
        output_cache = self.orig_fetch_from_cache(quant_cache, blocks, permutations)
        for i in range(len(output_cache)):
            output_cache[i] = self.quant_output(output_cache[i])
        return output_cache


class PatchedConv2d(nn.Conv2d):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input):
        qinput = self.quant_input(input)
        output = conv2d_fp8(
            qinput,
            self.weight,
            self.bias,
            self.stride,
            self.padding,
            self.dilation,
            self.groups,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        return output

    def forward_measure(self, input):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedSoftmax(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_output = self._mod_extra_config.outputs[0]
            # input scale is 1 assuming the input to SM is descaled because we are using HW supported scales
            self.scale_input = nn.Parameter(torch.Tensor([1.0]))
            self.scale_output = nn.Parameter(torch.Tensor([1 / mod_extra_config.scale.outputs[0]]))
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, x, dim=None, invAttnHead=None):
        output = torch.ops.hpu.softmax_fp8(x, dim, self.scale_input, self.scale_output, invAttnHead)
        return self.quant_output(output)

    def forward_measure(self, x, dim=None, invAttnHead=None):
        measure_input((x,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(x, dim, invAttnHead)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_output"),
        )


class PatchedLoRACompatibleLinear(nn.Linear):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.weight = nn.Parameter(self.weight.t().contiguous())
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input, scale: float = 1.0):
        qinput = self.quant_input(input)
        y = matmul_fp8(
            qinput,
            self.weight,
            out_dtype=self._mod_extra_config.config_params["hp_dtype"],
            scale_input_inv=self.scale_input,
            scale_other_inv=self.scale_weight,
        )
        output = y + self.bias if (self.bias is not None) else y
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
            # output = output + (scale * self.lora_layer(input))
        return output

    def forward_measure(self, input, scale: float = 1.0):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, scale)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedLoRACompatibleConv(nn.Conv2d):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_input = self._mod_extra_config.inputs[0]
            self.scale_input = nn.Parameter(mod_extra_config.scale.inputs[0])
            self.scale_weight = nn.Parameter(mod_extra_config.scale.params["weight"])
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(self, input, scale: float = 1.0):
        qinput = self.quant_input(input)
        if self.lora_layer is not None:
            # TODO SW-174899 support lora layer quantization
            _raise_lora_layer_error(self.class_name_org)
            # output = conv2d_fp8(qinput, self.weight, None, self.stride, self.padding, self.dilation, self.groups, \
            #  out_dtype=self._mod_extra_config.config_params["hp_dtype"], scale_input_inv=self.scale_input, scale_other_inv=self.scale_weight)
            # output = output + (scale * self.lora_layer(input))
            # output = output+torch.unsqueeze(torch.unsqueeze(self.bias,1), 1) if (self.bias is not None) else output
        else:
            output = conv2d_fp8(
                qinput,
                self.weight,
                self.bias,
                self.stride,
                self.padding,
                self.dilation,
                self.groups,
                out_dtype=self._mod_extra_config.config_params["hp_dtype"],
                scale_input_inv=self.scale_input,
                scale_other_inv=self.scale_weight,
            )
        return output

    def forward_measure(self, input, scale: float = 1.0):
        measure_input((input,), observer=self._mod_extra_config.inputs)
        output = self.orig_mod(input, scale)
        measure_output((output,), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(self, "scale_input", "scale_weight"),
        )


class PatchedModuleFusedSDPA(nn.Module):
    def __init__(self, mod, mod_extra_config, *args, **kwargs):
        # fsdpa is combined out of - BMM1(Q,K) -> Softmax -> BMM2(AMAX,V)
        # during measure we receive the amax value from the cguid and apply it during quant as input
        super().__init__()
        set_attrs_from_orig_model(self, mod, mod_extra_config)
        if self.quantization_mode == QuantMode.QUANTIZE:
            self.quant_q = self._mod_extra_config.inputs[0]
            self.quant_k = self._mod_extra_config.inputs[1]
            self.quant_v = self._mod_extra_config.inputs[2]
            self.dequant_output = self._mod_extra_config.outputs[0]
            self.scale_q = nn.Parameter(mod_extra_config.scale.inputs[0].type(torch.float32))
            self.scale_k = nn.Parameter(mod_extra_config.scale.inputs[1].type(torch.float32))
            self.scale_v = nn.Parameter(mod_extra_config.scale.inputs[2].type(torch.float32))
            self.descale_amax = nn.Parameter(mod_extra_config.scale.inputs[3].type(torch.float32))
            self.scale_output = nn.Parameter(1 / mod_extra_config.scale.outputs[0].type(torch.float32))
            self.scale_amax = nn.Parameter(1 / self.descale_amax)
        elif (self.quantization_mode == QuantMode.MEASURE) or (self.quantization_mode == QuantMode.SHAPE):
            self.forward = self.forward_measure

    def forward(
        self,
        q,
        k,
        v,
        attn_mask=None,
        dropout_p=0.0,
        is_causal=False,
        scale=None,
        softmax_mode="None",
    ):
        qinput = self.quant_q(q).detach()
        kinput = self.quant_k(k).detach()
        vinput = self.quant_v(v).detach()
        results = fp8_fused_sdpa(
            qinput,
            kinput,
            vinput,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
            # fp8_fused_sdpa in fp8 mode supports only FastSoftmax
            softmax_mode="None",
            d_scale_q=self.scale_q,
            d_scale_k=self.scale_k,
            d_scale_v=self.scale_v,
            q_scale_s=self.scale_amax,
            q_scale_o=self.scale_output,
            d_scale_s=self.descale_amax,
            is_amax_s=False,
        )
        output = results[0]
        d_out = self.dequant_output(output)
        return d_out

    def forward_measure(
        self,
        q,
        k,
        v,
        attn_mask=None,
        dropout_p=0.0,
        is_causal=False,
        scale=None,
        softmax_mode="fast",
    ):
        dq = q.detach()
        dk = k.detach()
        dv = v.detach()
        measure_input((dq, dk, dv), observer=self._mod_extra_config.inputs)
        results = fp8_fused_sdpa(
            dq,
            dk,
            dv,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
            # fp8_fused_sdpa in bf16 can use either FastSoftmax or regular
            softmax_mode="fast",
            is_amax_s=True,
        )
        output = results[0]
        amax = results[1]
        measure_output((output, amax), self._mod_extra_config.outputs)
        return output

    def extra_repr(self) -> str:
        return extra_representation(
            self.extra_repr_org(),
            self.class_name_org,
            get_current_repr(
                self,
                "scale_q",
                "scale_k",
                "scale_v",
                "descale_amax",
                "scale_amax",
                "scale_output",
            ),
        )


class PatchedUnmeasuredModule(nn.Module):
    def __init__(self, name, *args, **kwargs):
        super().__init__()
        self.name = name

    def forward(self, *args, **kwargs):
        raise Exception(
            "Error - Layer '{}' was called but was not quantized because no measures were supplied.".format(self.name)
        )

    def extra_repr(self) -> str:
        return f"Dummy patch of {self.name} to raise exception as there are no measurements provided."



# -*- coding: utf-8 -*-
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch environment check."""

import importlib
import sys

import torch
from packaging.version import Version


################ Check imported sys.module first to decide behavior #################
def is_ipex_imported() -> bool:
    """Check whether intel_extension_for_pytorch is imported."""
    for name, _ in sys.modules.items():
        if name == "intel_extension_for_pytorch":
            return True
    return False


def is_transformers_imported() -> bool:
    """Check whether transformers is imported."""
    for name, _ in sys.modules.items():
        if name == "transformers":
            return True
    return False


################ Check available sys.module to decide behavior #################
def is_package_available(package_name):
    """Check if the package exists in the environment without importing.

    Args:
        package_name (str): package name
    """
    from importlib.util import find_spec

    package_spec = find_spec(package_name)
    return package_spec is not None


## check hpex
if is_package_available("habana_frameworks"):
    _hpex_available = True
    import habana_frameworks.torch.hpex  # pylint: disable=E0401
else:
    _hpex_available = False


def is_hpex_available():
    """Returns whether hpex is available."""
    return _hpex_available


## check ipex
if is_package_available("intel_extension_for_pytorch"):
    _ipex_available = True
else:
    _ipex_available = False


def is_ipex_available():
    """Return whether ipex is available."""
    return _ipex_available


def get_ipex_version():
    """Return ipex version if ipex exists."""
    if is_ipex_available():
        try:
            import intel_extension_for_pytorch as ipex

            ipex_version = ipex.__version__.split("+")[0]
        except ValueError as e:  # pragma: no cover
            assert False, "Got an unknown version of intel_extension_for_pytorch: {}".format(e)
        version = Version(ipex_version)
        return version
    else:
        return None


TORCH_VERSION_2_2_2 = Version("2.2.2")


def get_torch_version():
    """Return torch version if ipex exists."""
    try:
        torch_version = torch.__version__.split("+")[0]
    except ValueError as e:  # pragma: no cover
        assert False, "Got an unknown version of torch: {}".format(e)
    version = Version(torch_version)
    return version


GT_TORCH_VERSION_2_3_2 = get_torch_version() > Version("2.3.2")


def get_accelerator(device_name="auto"):
    """Return the recommended accelerator based on device priority."""
    global accelerator  # update the global accelerator when calling this func
    from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

    accelerator = auto_detect_accelerator(device_name)
    return accelerator


# for direct user access, used by @device_synchronize, can be changed by set_accelerator
accelerator = get_accelerator()


# for habana ease-of-use
def device_synchronize(raw_func):
    """Function decorator that calls accelerated.synchronize before and after a function call."""
    from functools import wraps

    @wraps(raw_func)
    def new_func(*args, **kwargs):
        accelerator.synchronize()
        output = raw_func(*args, **kwargs)
        accelerator.synchronize()
        return output

    return new_func



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import habana_frameworks.torch.core as htcore
import torch
import torch.nn as nn

from .._quant_common.helper_modules import PatchedUnmeasuredModule
from .._quant_common.quant_config import get_hqt_config
from ..utils.logger import logger
from .common import UNMEASURED_MODELS, generate_model_info, mod_default_dict, parent_child_mod_dict
from .measure import load_measurements
from .scale import get_config, scale_method_mapping, scaling_methods


def patch_module(mod, qconfig, mod_dict, patched_mod=None):
    parent = parent_child_mod_dict[mod].parent
    name = parent_child_mod_dict[mod].name
    if patched_mod is None:
        patched_mod = mod_dict[mod.__class__.__name__].patched_module(mod, qconfig)
    setattr(parent, name, patched_mod)


def apply_hf_hook(module):
    if hasattr(module, "_hf_hook"):
        module._hf_hook.pre_forward(module)
        module._hf_hook.detach_hook(module)
        delattr(module, "_hf_hook")
    if hasattr(module, "_old_forward"):
        module.forward = module._old_forward
        delattr(module, "_old_forward")


def quantize_params(mod, mod_extra_config):
    for param_name in mod_extra_config.params:
        quantizer = mod_extra_config.params[param_name]
        param = getattr(mod, param_name)
        quantized_param = quantizer(param.to("hpu"))
        delattr(mod, param_name)
        setattr(mod, param_name, nn.Parameter(quantized_param))
        quantized_param = getattr(mod, param_name)
        quantized_param.requires_grad_(False)
        htcore.mark_step()


def prepare_model(model, qconfig, mod_list, hp_dtype=torch.float):
    config = get_hqt_config(model)
    patched_modules = []
    patched_module_types = set()
    with torch.no_grad():
        for name, mod in model.named_modules():
            if name in qconfig[UNMEASURED_MODELS]:
                if not config.cfg["ignore_modules_wo_measures"]:
                    patch_module(mod, None, None, PatchedUnmeasuredModule(name))
                else:
                    logger.debug("Module %s was not quantized.", name)
                continue
            # When offloading weight to disk, need to transfer the weight from disk to cpu using hf_hook
            apply_hf_hook(mod)
            if name in mod_list:
                mod_extra_config = qconfig[name]
                quantize_params(mod, mod_extra_config)
                patch_module(mod, mod_extra_config, mod_default_dict)
                patched_modules.append(name)
                patched_module_types.add(type(mod))
    logger.debug("Patched module types: %s", patched_module_types)
    logger.debug("Patched modules: %s", patched_modules)
    logger.debug("Total patched modules: %d", len(patched_modules))
    model = model.to("hpu")
    htcore.mark_step()


def quantize(model, mod_list):
    config = get_hqt_config(model)
    generate_model_info(model)
    hp_dtype = config.cfg["hp_dtype"]
    lp_dtype = config.cfg["fp8_config"]
    measurement = load_measurements(model, config.cfg["measure_file"])
    # FIXME make sure this takes unit_scale or measured scale, from Configs
    scaling_method_name = scale_method_mapping[(config.cfg["scale_method"], config.cfg["observer"])]
    scaling_method = scaling_methods[scaling_method_name]
    params = config.cfg["scale_params"]
    params["hp_dtype"] = hp_dtype
    params["lp_dtype"] = lp_dtype
    qconfig = get_config(
        model,
        measurement,
        mod_default_dict,
        scaling_method,
        params,
        config.cfg["scale_file"],
        False,
        mod_list,
    )
    prepare_model(model, qconfig, mod_list, hp_dtype=hp_dtype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GPTQ quantization."""

import gc
import math
import random
import re
import time
from collections import UserDict, defaultdict
from functools import partial

import torch
import torch.nn as nn
from tqdm import tqdm

from neural_compressor.torch.utils import (
    get_accelerator,
    get_model_device,
    is_transformers_imported,
    logger,
    set_module,
)
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

from .modules import INCWeightOnlyLinear

if is_transformers_imported():
    import transformers

    SUPPORTED_LAYERS = [nn.Conv2d, nn.Conv1d, nn.Linear, transformers.Conv1D]
else:
    SUPPORTED_LAYERS = [nn.Conv2d, nn.Conv1d, nn.Linear]
DEBUG = False
accelerator = auto_detect_accelerator()


# ==============model structure related==============
def is_leaf(module):
    """Judge whether a module has no child-modules.

    Args:
        module: torch.nn.Module

    Returns:
        a bool: whether a module has no child-modules.
    """
    children_cnt = 0
    for n in module.children():
        children_cnt += 1
    return True if children_cnt == 0 else False


def trace_gptq_target_blocks(module, module_types=[torch.nn.ModuleList, torch.nn.Sequential]):
    """Search transformer stacked structures, which is critical in LLMs and GPTQ execution.

    Args:
        module: torch.nn.Module
        module_types: List of torch.nn.Module.

    Returns:
        gptq_related_blocks = {
            "embeddings": {}, # Dict embedding layers before transformer stack module,
            "transformers_pre": {}, # TODO
            "transformers_name": string. LLMs' transformer stack module name ,
            "transformers": torch.nn.ModuleList. LLMs' transformer stack module,
            "transformers": {}, Dict# TODO
        }
    """
    if type(module).__name__ == "MixFormerSequentialForCausalLM":  # pragma: no cover
        gptq_related_blocks = {
            "embeddings": {},
            "transformers_pre": {},  # todo
            "transformers_name": "",  # None
            "transformers": [],  # None
            "transformers_post": {},  # todo
        }
        for n, m in module.named_modules():
            if type(m) in module_types:
                gptq_related_blocks["transformers_name"] = n
                gptq_related_blocks["transformers"] = m
                break
            else:
                continue
        for n, m in gptq_related_blocks["transformers"][0].named_modules():
            if is_leaf(m):
                gptq_related_blocks["embeddings"][n] = m
        gptq_related_blocks["transformers"] = gptq_related_blocks["transformers"][1:-1]
    else:
        gptq_related_blocks = {
            "embeddings": {},
            "transformers_pre": {},  # todo
            "transformers_name": "",  # None
            "transformers": [],  # None
            "transformers_post": {},  # todo
        }
        for n, m in module.named_modules():
            if type(m) in module_types:
                gptq_related_blocks["transformers_name"] = n
                gptq_related_blocks["transformers"] = m
                return gptq_related_blocks
            else:
                if is_leaf(m):
                    gptq_related_blocks["embeddings"][n] = m
    return gptq_related_blocks


def find_layers(module, layers=SUPPORTED_LAYERS, name=""):
    """Get all layers with target types."""
    if type(module) in layers:
        return {name: module}
    else:
        # use string type to find name:
        if isinstance(module, tuple(layers)):
            return {name: module}
        else:
            pass
    res = {}
    for name1, child in module.named_children():
        res.update(find_layers(child, layers=layers, name=name + "." + name1 if name != "" else name1))
    return res


def find_layers_name(module, layers=SUPPORTED_LAYERS, name=""):
    """Get all layers with target types."""
    if type(module) in layers:
        return [name]
    res = []
    for name1, child in module.named_children():
        res += find_layers_name(child, layers=layers, name=name + "." + name1 if name != "" else name1)
    return res


def log_quantizable_layers_per_transformer(transformer_blocks, layers=SUPPORTED_LAYERS):
    """Print all layers which will be quantized in GPTQ algorithm."""
    logger.info("* * Layer to be quantized * *")

    for block_id in range(len(transformer_blocks["transformers"])):
        transformer_block = transformer_blocks["transformers"][block_id]
        layers_for_this_tblock = find_layers_name(transformer_block)
        layer_names = [
            (transformer_blocks["transformers_name"] + "." + str(block_id) + "." + layer_name)
            for layer_name in layers_for_this_tblock
        ]
        for name in layer_names:
            logger.info(name)


class RAWGPTQuantizer(object):
    """Main API for GPTQ algorithm.

    Please refer to:
    GPTQ: Accurate Post-training Compression for Generative Pretrained Transformers
    url: https://arxiv.org/abs/2210.17323
    """

    def __init__(
        self,
        model,
        weight_config={},
        nsamples=128,
        use_max_length=True,
        max_seq_length=2048,
        device=None,
        use_layer_wise=False,
        model_path="",
        dataloader=None,
        *args,
        **kwargs,
    ):
        """Init RAWGPTQuantizer.

        Args:
            model: the fp32 model to quantize
            weight_config (dict, optional): contains all info required by GPTQ. Defaults to {}. For example,
            weight_config={
                'layer1':
                {
                    'bits': 4,
                    'group_size': 32,
                    'sym': False,
                    'percdamp': .01,
                    'act_order': False
                }
                ...
            }
            nsamples (int): the number of calibration data samples.
            use_max_length (bool): set all sequence length to be same length.
            max_seq_length (int): the same length of all sequence length.
            dataloader: an iterable containing calibration datasets, contains (inputs, targets)
            use_layer_wise (bool): Enables quantize model per layer. Defaults to False.
            model_path (str): Model path that is used to load state_dict per layer.
            device (str): cpu or cuda.
        """
        # model
        self.model = model
        # self.use_cache = self.model.config.use_cache
        self.gptq_related_blocks = trace_gptq_target_blocks(self.model)  # get the transformer block list above
        self.dtype = next(iter(self.model.parameters())).dtype
        log_quantizable_layers_per_transformer(self.gptq_related_blocks)

        # weight config
        self.weight_config = weight_config
        # default settings, check configs
        self.dtype_default = "int"
        self.bits_default = 4
        self.group_size_default = 128
        self.block_size_default = 128
        self.percdamp_default = 0.01
        self.sym_default = False
        self.act_order_default = False
        self.static_groups_default = False
        self.perchannel_default = True
        self.mse_default = False
        self.use_double_quant_default = False
        self.double_quant_dtype_default = "int"
        self.double_quant_bits_default = 4
        self.double_quant_group_size_default = 128
        self.double_quant_sym_default = False
        self.check_layer_config()

        # device
        self.device = get_accelerator(kwargs.pop("device", "auto")).current_device_name()
        if not use_layer_wise:
            self.model.to(self.device)
        self.is_ready = False

        self.use_layer_wise = use_layer_wise
        if use_layer_wise:
            self.prepare_layer_wise(model_path)

        # dataloader
        self.use_max_length = use_max_length
        self.max_seq_length = max_seq_length
        self.dataloader_original = dataloader
        self.dataloader = []
        self.nsamples = nsamples

    def prepare_layer_wise(self, model_path):
        """Prepare for layer-wise quantization, including registering hooks and setting up the model path.

        Args:
            model_path (str): Model path that is used to load state_dict per layer.
        """
        import os

        from neural_compressor.torch.algorithms.layer_wise import LWQ_WORKSPACE, get_path, register_weight_hooks

        os.makedirs(LWQ_WORKSPACE, exist_ok=True)
        if model_path == "":
            model_path = self.model.path
        assert model_path, "model_path should not be None."
        self.model_path = get_path(model_path)
        register_weight_hooks(
            self.model, self.model_path, device=self.device, clean_weight=True, saved_path=LWQ_WORKSPACE
        )

    def get_full_layer_name(self, sub_layer_name, block_idx):
        """Get full layer name.

        Args:
            sub_layer_name (str): sub layer name
            block_idx (int): index of block

        Returns:
            str: The full name of layer.
        """
        transformer_name = self.gptq_related_blocks["transformers_name"]
        return ".".join([transformer_name, str(block_idx), sub_layer_name])

    def check_layer_config(self):
        """Copy arguments from weight_config to built-in attributes."""
        for layer_name, config in self.weight_config.items():
            self.weight_config[layer_name]["dtype"] = config.get("dtype", self.dtype_default)
            self.weight_config[layer_name]["bits"] = config.get("bits", self.bits_default)
            self.weight_config[layer_name]["group_size"] = config.get("group_size", self.group_size_default)
            self.weight_config[layer_name]["block_size"] = config.get("block_size", self.group_size_default)
            self.weight_config[layer_name]["percdamp"] = config.get("percdamp", self.percdamp_default)
            self.weight_config[layer_name]["sym"] = config.get("sym", self.sym_default)
            self.weight_config[layer_name]["act_order"] = config.get("act_order", self.act_order_default)
            self.weight_config[layer_name]["static_groups"] = config.get("static_groups", self.static_groups_default)
            self.weight_config[layer_name]["perchannel"] = config.get("perchannel", self.perchannel_default)
            self.weight_config[layer_name]["mse"] = config.get("mse", self.mse_default)
            self.weight_config[layer_name]["use_double_quant"] = config.get(
                "use_double_quant", self.use_double_quant_default
            )
            self.weight_config[layer_name]["double_quant_dtype"] = config.get(
                "double_quant_dtype", self.double_quant_dtype_default
            )  # only support int
            self.weight_config[layer_name]["double_quant_bits"] = config.get(
                "double_quant_bits", self.double_quant_bits_default
            )
            self.weight_config[layer_name]["double_quant_group_size"] = config.get(
                "double_quant_group_size", self.double_quant_group_size_default
            )
            self.weight_config[layer_name]["double_quant_sym"] = config.get(
                "double_quant_sym", self.double_quant_sym_default
            )
            if self.weight_config[layer_name]["dtype"] != "int" and "int" in self.weight_config[layer_name]["dtype"]:
                self.weight_config[layer_name]["bits"] = int(self.weight_config[layer_name]["dtype"].lstrip("int"))
                self.weight_config[layer_name]["dtype"] = "int"

    def get_layer_config(self, layer_name):
        """Obtain config for one layer, since GPTQ supports layer-wise config."""
        # First try the exact name matching, if cannot find, use re to search. For example, can support ".*" in op_name
        config = None
        config = self.weight_config.get(layer_name, None)
        if config is not None:
            return config
        else:
            for k, v in self.weight_config.items():
                regex = re.compile(k)
                if len(regex.findall(layer_name)) is not None:
                    config = v
                    return config
                else:
                    pass
        return config

    def track_hidden_states(self, data):
        """Track hidden states.

        Args:
            data (tensor/tuple/list): input data.

        Returns:
            tensor.
        """
        if isinstance(data, torch.Tensor):
            return data
        elif isinstance(data, tuple) or isinstance(data, list):
            return data[0]

    @torch.no_grad()
    def prepare_for_calibration(self):
        """Prepare input calibration data and other attributes which are critical for gptq execution."""
        try:
            self.cache_key_arguments = {
                "batch_num": 0
            }  # a dict of list, keyword arguments ("attention_masks", "position_ids", etc.)
            # Note that the first elements in cache_positional_arguments is main input: hidden_states
            self.cache_positional_arguments = []  # a list of list, positional arguments ("rotary_pos_emb" in chatglm)
            self.is_ready = True
        except:
            logger.warning("GPTQ Quantizer initialization failed!")
            pass

        # critical: hooker function which collects inputs
        def forward(layer, *args, **kwargs):
            # inputs[inputs_info['idx']] = input_ids # TODO solve the problem of batchsize!=1
            self.cache_key_arguments["batch_num"] += 1
            for arg in kwargs:
                # TODO: investigate include parameters
                # each outputs can be different shape, hence also use list to store
                if isinstance(kwargs[arg], torch.Tensor) or arg == "alibi":
                    if self.cache_key_arguments.get(arg, None) is None:
                        self.cache_key_arguments[arg] = []
                    self.cache_key_arguments[arg].append(kwargs[arg])
                continue
            # copy positional arguments, positional arguments are sensitive for their order, be cautious!
            # Most models in HF has avoid this, but some models still use positional arguments other than
            # hidden_states, chatglm2-6b etc.
            for idx, item in enumerate(args):
                if (idx + 1) > len(self.cache_positional_arguments):
                    # initialize
                    self.cache_positional_arguments.append([])
                self.cache_positional_arguments[idx].append(item)
            raise ValueError

        # Step1: fetch the embeddings and other layers before the transformer stack.
        if not self.use_layer_wise:  # pragma: no cover
            for embedding_name, embedding_layer in self.gptq_related_blocks["embeddings"].items():
                embedding_layer = embedding_layer.to(self.device)

        # Step2: modify the first transformer block's forward function to obtain inputs for calibration
        if not self.use_layer_wise:  # pragma: no cover
            self.gptq_related_blocks["transformers"][0] = self.gptq_related_blocks["transformers"][0].to(self.device)
        self.forward_cache = self.gptq_related_blocks["transformers"][0].forward
        self.gptq_related_blocks["transformers"][0].forward = partial(
            forward, self.gptq_related_blocks["transformers"][0]
        )
        # Step 3: replace model_forward to avoid ValueError
        self.orig_model_forward_cache = self.model.forward
        model_forward_cache = self.model.forward

        def model_forward(model, *args, **kwargs):
            nonlocal model_forward_cache
            try:
                model_forward_cache(*args, **kwargs)
            except ValueError:
                pass

        self.model.forward = partial(model_forward, self.model)

    @torch.no_grad()
    def remove_prepare_for_calibration(self):
        """Prepare for GPTQ quantization."""
        # output inp data shape
        logger.info("All calibration data's shape =>")
        # check all hidden_states shape
        try:
            for hidden_states in self.cache_positional_arguments[0]:
                logger.info(hidden_states.shape)
        except:
            pass
        logger.info("Done.")

        # Step 4: restore original forward function, relocate layers back to cpu.
        self.model.forward = self.orig_model_forward_cache
        self.gptq_related_blocks["transformers"][0].forward = self.forward_cache
        if not self.use_layer_wise:  # pragma: no cover
            self.gptq_related_blocks["transformers"][0] = self.gptq_related_blocks["transformers"][0].cpu()
            for embedding_name, embedding_layer in self.gptq_related_blocks["embeddings"].items():
                embedding_layer.to(self.device)
        torch.cuda.empty_cache()
        # end
        logger.info("GPTQ quantization prepared.")

    def gather_single_batch_from_dict(self, data_dict, idx):
        """Gather single batch from a dict.

        Args:
            data_dict (dict): data dict.
            idx (int): index

        Returns:
            dict: single batch.
        """
        # obtain a set of keyword input from cache
        single_batch = {}
        for k, v in data_dict.items():
            single_batch[k] = data_dict[k][idx]
        return single_batch

    def gather_single_batch_from_list(self, data_list, idx):
        """Gather single batch from a list.

        Args:
            data_dict (dict): data list.
            idx (int): index

        Returns:
            list: single batch.
        """
        # obtain a set of keyword input from cache
        single_batch = []
        for data_item in data_list:
            single_batch.append(data_item[idx])
        return single_batch

    def update_blockwise_hidden_states(self, outs):
        """Update the blockwise hidden states.

        Args:
            outs: the output of block.
        """
        if "hidden_states" in self.cache_key_arguments:
            self.cache_key_arguments["hidden_states"] = outs[:]
        else:
            self.cache_positional_arguments[0] = outs[:]

    @torch.no_grad()
    def execute_quantization(self, means=None, stds=None):
        """Run quantization."""
        # Step1: prepare quantization (calibration datasets)

        logger.info("Begin ====>")

        # Step2: run gptq quantization in a transformer block-wise manner.
        gptq_config = {}
        tblock_length = len(self.gptq_related_blocks["transformers"])
        for block_idx in range(tblock_length):
            logger.info(f"Quantizing layer {block_idx + 1} / {tblock_length}..")
            if not self.use_layer_wise:  # pragma: no cover
                # if we do not apply layer-wise feature, we still place the entire block on the GPU
                transformer_block = self.gptq_related_blocks["transformers"][block_idx].to(self.device)
            else:
                transformer_block = self.gptq_related_blocks["transformers"][block_idx]  # .to(self.device)
            # Step2.1: obtain all layers (Linear, Conv2d, etc) in the block which can be quantized.
            sub_layers = find_layers(transformer_block)
            sub_layers_to_quant = {}
            for layer_name, layer_obj in sub_layers.items():
                # filter sub_layers with included layer_names in self.weight_config
                full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                # if self.weight_config.get(full_layer_name, None) == None:
                if self.get_layer_config(full_layer_name) is None:
                    logger.warning(f"{full_layer_name} can be quantized " + "but excluded from quantization configs.")
                else:
                    sub_layers_to_quant[layer_name] = layer_obj
            del sub_layers
            sub_layers = sub_layers_to_quant
            # Step 2.2: Initialize GPTQ quantizers for collected layers.
            gptq_for_this_block = {}
            # initialize gptq quantizer for every layer in a transformer block
            for layer_name in sub_layers:
                # weight_config_this_layer = self.weight_config.get(
                #     self.get_full_layer_name(layer_name, block_idx), None
                # )
                full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                weight_config_this_layer = self.get_layer_config(full_layer_name)
                if self.use_layer_wise:  # pragma: no cover
                    from neural_compressor.torch.algorithms.layer_wise import load_value

                    W = load_value(self.model, full_layer_name + ".weight", self.model_path)
                else:
                    W = sub_layers[layer_name].weight.data.clone()

                gptq_for_this_block[layer_name] = GPTQ(sub_layers[layer_name], W, self.device)
                # gptq_for_this_block[layer_name].quantizer = Quantizer()
                gptq_for_this_block[layer_name].quantizer.configure(weight_config_this_layer)

            # Step 2.3: modify forward functions to hook inputs data (used in gptq execution)
            def add_batch(_name):
                def tmp(_, inp, out):
                    gptq_for_this_block[_name].add_batch(inp[0].data, out.data)  # noqa: F821

                return tmp

            handles = []  # register handles which add inputs and outputs to gptq object
            for layer_name in sub_layers:
                handles.append(sub_layers[layer_name].register_forward_hook(add_batch(layer_name)))
            batch_num = self.cache_key_arguments.pop("batch_num")
            for j in range(batch_num):
                cache_keyword_batch = self.gather_single_batch_from_dict(self.cache_key_arguments, j)
                cache_positional_batch = self.gather_single_batch_from_list(self.cache_positional_arguments, j)
                accelerator.mark_step()
                out = transformer_block(*cache_positional_batch, **cache_keyword_batch)
                out = self.track_hidden_states(out)
            self.cache_key_arguments["batch_num"] = batch_num
            for h in handles:
                h.remove()
            # Step 2.4: everything is prepared, so start quantization!
            for layer_name in sub_layers:
                # weight_config_this_layer = self.weight_config.get(
                #     self.get_full_layer_name(layer_name, block_idx), None
                # )
                weight_config_this_layer = self.get_layer_config(self.get_full_layer_name(layer_name, block_idx))
                logger.info(f"Quantizing layer {layer_name}")
                if self.use_layer_wise:  # pragma: no cover
                    from neural_compressor.torch.algorithms.layer_wise import load_value

                    full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                    W = load_value(self.model, full_layer_name + ".weight", self.model_path)
                else:
                    W = sub_layers[layer_name].weight.data.clone()
                accelerator.mark_step()
                if "hpu" in self.device:
                    W = W.to("cpu")
                scale, zp, Q = gptq_for_this_block[layer_name].fasterquant(
                    W,
                    blocksize=weight_config_this_layer["block_size"],
                    percdamp=weight_config_this_layer["percdamp"],
                    groupsize=weight_config_this_layer["group_size"],
                    act_order=weight_config_this_layer["act_order"],
                    static_groups=weight_config_this_layer["static_groups"],
                )
                if self.use_layer_wise:  # pragma: no cover
                    from neural_compressor.torch.algorithms.layer_wise import (
                        LWQ_WORKSPACE,
                        clean_module_weight,
                        load_value,
                        set_module_tensor_to_device,
                    )

                    sub_layer = sub_layers[layer_name]
                    full_layer_name = self.get_full_layer_name(layer_name, block_idx)
                    for n, p in sub_layer.named_parameters():
                        param_name = full_layer_name + "." + n
                        if n == "weight":
                            set_module_tensor_to_device(self.model, param_name, self.device, Q)
                        else:
                            value = load_value(self.model, param_name, self.model_path)
                            set_module_tensor_to_device(self.model, param_name, self.device, value)
                    # sub_layer.weight.data = Q
                    torch.save(sub_layer.state_dict(), LWQ_WORKSPACE + f"/{full_layer_name}.pt")
                    clean_module_weight(sub_layer)
                    del Q
                    gc.collect()
                else:
                    sub_layers[layer_name].weight.data = Q
                gptq_config[self.get_full_layer_name(layer_name, block_idx)] = {"scale": scale}
                if not weight_config_this_layer["sym"]:
                    gptq_config[self.get_full_layer_name(layer_name, block_idx)]["zero"] = zp
                if weight_config_this_layer["act_order"]:  # save perm for restoring the weights
                    gptq_config[self.get_full_layer_name(layer_name, block_idx)]["perm"] = gptq_for_this_block[
                        layer_name
                    ].perm
                gptq_for_this_block[layer_name].free()

            # Step 2.5: replace output data with quantized weights
            outs = []
            batch_num = self.cache_key_arguments.pop("batch_num")
            for j in range(batch_num):
                cache_keyword_batch = self.gather_single_batch_from_dict(self.cache_key_arguments, j)
                cache_positional_batch = self.gather_single_batch_from_list(self.cache_positional_arguments, j)
                out = transformer_block(*cache_positional_batch, **cache_keyword_batch)
                out = self.track_hidden_states(out)
                outs.append(out)
            self.cache_key_arguments["batch_num"] = batch_num
            if self.use_layer_wise:  # pragma: no cover
                self.gptq_related_blocks["transformers"][block_idx] = transformer_block
            else:
                self.gptq_related_blocks["transformers"][block_idx] = transformer_block.cpu()
            # Step 2.6: export to compressed model
            for layer_name in sub_layers:
                weight_config_this_layer = self.get_layer_config(self.get_full_layer_name(layer_name, block_idx))
                gptq_scale = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["scale"]
                if not weight_config_this_layer["sym"]:
                    gptq_zp = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["zero"]
                else:
                    gptq_zp = None
                if weight_config_this_layer["act_order"]:  # save perm for restoring the weights
                    gptq_perm = gptq_config[self.get_full_layer_name(layer_name, block_idx)]["perm"]
                else:
                    gptq_perm = None
                if self.use_layer_wise:
                    state_dict = torch.load(LWQ_WORKSPACE + f"/{self.get_full_layer_name(layer_name, block_idx)}.pt")
                    Q = state_dict["weight"].data
                    bias = state_dict["bias"] if "bias" in state_dict.keys() else None

                else:
                    Q = sub_layers[layer_name].weight.data
                if weight_config_this_layer["act_order"]:
                    Q.copy_(Q[:, gptq_perm])
                if is_transformers_imported() and isinstance(sub_layers[layer_name], transformers.Conv1D):
                    Q = Q.t_().contiguous()
                from .utility import quant_weight_w_scale

                quant_weight_w_scale(
                    Q,
                    gptq_scale,
                    gptq_zp,
                    weight_config_this_layer["group_size"],
                    dtype=weight_config_this_layer["dtype"],
                )
                if weight_config_this_layer["act_order"]:
                    invperm = torch.argsort(gptq_perm)
                    Q.copy_(Q[:, invperm])
                int_weight = Q.type(torch.int32)  # copy_ is not workable for different types.
                # replace module
                if isinstance(sub_layers[layer_name], torch.nn.Linear):
                    in_features = sub_layers[layer_name].in_features
                    out_features = sub_layers[layer_name].out_features
                elif is_transformers_imported() and isinstance(sub_layers[layer_name], transformers.Conv1D):
                    in_features = sub_layers[layer_name].weight.shape[0]
                    out_features = sub_layers[layer_name].weight.shape[1]
                    int_weight = sub_layers[layer_name].weight.t_().contiguous()
                    scale = scale.t_().contiguous()
                    zp = zp.t_().contiguous() if zp is not None else zp

                if not self.use_layer_wise:
                    bias = sub_layers[layer_name].bias

                new_module = INCWeightOnlyLinear(
                    in_features,
                    out_features,
                    dtype=weight_config_this_layer["dtype"],
                    bits=weight_config_this_layer["bits"],
                    group_size=weight_config_this_layer["group_size"],
                    zp=gptq_zp is not None,
                    bias=bias is not None,
                    g_idx=gptq_perm is not None,
                    device=self.device,
                )
                new_module.pack(int_weight, gptq_scale, gptq_zp, bias, gptq_perm)
                set_module(transformer_block, layer_name, new_module)
            del gptq_for_this_block
            torch.cuda.empty_cache()
            # iteratively replace the input with output, thus layerwise quantization can continue.
            self.update_blockwise_hidden_states(outs)
            logger.info("------------------------------")

        logger.info("Quantization done")
        # self.model.config.use_cache = self.use_cache

        # obtain model (all weight only quantization API function should return)
        for k, v in gptq_config.items():
            for m, n in v.items():
                gptq_config[k][m] = n.tolist()
        return self.model, gptq_config


class GPTQ:
    """Please refer to the following.

    GPTQ: Accurate Post-training Compression for Generative Pretrained Transformers (https://arxiv.org/abs/2210.17323)
    """

    def __init__(self, layer, W, device="cpu"):
        """Init GPTQ."""
        self.layer = layer
        self.device = device
        # W = layer.weight.data.clone()
        if isinstance(self.layer, nn.Conv2d) or isinstance(self.layer, nn.Conv1d):
            W = W.flatten(1)
        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        self.rows = W.shape[0]  # output channels
        self.columns = W.shape[1]  # input channels
        self.H = torch.zeros((self.columns, self.columns), device=self.device)
        self.nsamples = 0
        self.quantizer = Quantizer()
        self.perm = None  # act_order choice

    def add_batch(self, inp, out):
        """Add inputs and outputs to gptq object."""
        # if DEBUG:
        #     self.inp1 = inp
        #     self.out1 = out
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear) or (
            is_transformers_imported() and isinstance(self.layer, transformers.Conv1D)
        ):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()
        # TODO: llm's transformer sequential with nn.conv2d is currently not under test
        # if isinstance(self.layer, nn.Conv2d):
        #     unfold = nn.Unfold(
        #         self.layer.kernel_size,
        #         dilation=self.layer.dilation,
        #         padding=self.layer.padding,
        #         stride=self.layer.stride
        #     )
        #     inp = unfold(inp)
        #     inp = inp.permute([1, 0, 2])
        #     inp = inp.flatten(1)
        self.H *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp
        # inp = inp.float()
        inp = math.sqrt(2 / self.nsamples) * inp.float()
        # self.H += 2 / self.nsamples * inp.matmul(inp.t())
        self.H += inp.matmul(inp.t())  # H = X*X, which should be a sym matrix

    def fasterquant(self, W, blocksize=128, percdamp=0.01, groupsize=-1, act_order=False, static_groups=False):
        """Run quantization.

        Args:
            W (tensor): weight tensor.
            block_size (int): Execute quantization per block, block shape = [C_out, block_size]. Default to 128.
            percdamp (float): percdamp (float): Percentage of Hessian's diagonal values' average, which will be added
                                to Hessian's diagonal to increase numerical stability. Defaults to 0.01.
            groupsize (int): Size of weight groups. Defaults to -1.
            act_order (bool): Whether to sort Hessian's diagonal values to rearrange channel-wise quantization order.
                                Defaults to False.
            static_groups (bool): Whether to calculate group wise quantization parameters in advance. This option
                                    mitigate actorder's extra computational requirements. Default to False.

        Returns:
            scale, zero, Q
        """
        # W = self.layer.weight.data.clone()
        weight_shape, weight_dtype = W.shape, W.data.dtype
        if isinstance(self.layer, nn.Conv2d):
            W = W.flatten(1)
        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            W = W.t()
        W = W.float()

        tick = time.time()

        if not self.quantizer.ready():
            self.quantizer.find_params(W, weight=True)

        H = self.H
        if "hpu" in self.device:
            H = H.to("cpu")
        del self.H
        dead = torch.diag(H) == 0
        H[dead, dead] = 1
        W[:, dead] = 0  # such channel makes no contribution to quantization computation

        # enable static_groups
        # calculate the quantization parameters for original group in advance.
        if static_groups:
            import copy

            groups = []
            for i in range(0, self.columns, groupsize):
                quantizer = copy.deepcopy(self.quantizer)
                quantizer.find_params(W[:, i : (i + groupsize)], weight=True)
                groups.append(quantizer)

        # rearrange considering the diag's value
        if act_order:
            perm = torch.argsort(torch.diag(H), descending=True)
            W = W[:, perm]
            H = H[perm][:, perm]
            self.perm = perm.clone()

        Losses = torch.zeros_like(W)
        Q = torch.zeros_like(W)

        damp = percdamp * torch.mean(torch.diag(H))
        diag = torch.arange(self.columns, device=self.device)
        H[diag, diag] += damp  # add a average value of
        H = torch.linalg.cholesky(H)
        H = torch.cholesky_inverse(H)
        H = torch.linalg.cholesky(H, upper=True)
        Hinv = H

        scale = []
        zero = []

        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1

            W1 = W[:, i1:i2].clone()
            Q1 = torch.zeros_like(W1)
            Err1 = torch.zeros_like(W1)
            Losses1 = torch.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            for i in range(count):  # within a block, channel wise
                w = W1[:, i]
                d = Hinv1[i, i]

                if groupsize != -1:
                    if not static_groups:
                        if (i1 + i) % groupsize == 0:
                            self.quantizer.find_params(W[:, (i1 + i) : (i1 + i + groupsize)], weight=True)
                            scale.append(self.quantizer.scale)
                            zero.append(self.quantizer.zero)
                    else:
                        idx = i1 + i
                        if act_order:
                            idx = perm[idx]
                        self.quantizer = groups[idx // groupsize]
                q = self.quantizer.quantize(
                    w.unsqueeze(1), self.quantizer.scale, self.quantizer.zero, self.quantizer.maxq
                ).flatten()
                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d**2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1

            Q[:, i1:i2] = Q1
            Losses[:, i1:i2] = Losses1 / 2

            W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])

            # if DEBUG:
            #     self.layer.weight.data[:, :i2] = Q[:, :i2]
            #     self.layer.weight.data[:, i2:] = W[:, i2:]
            #     logger.info(f"{torch.sum((self.layer(self.inp1) - self.out1) ** 2)}")
            #     logger.info(f"{torch.sum(Losses)}")

        if str(self.device).startswith("cuda"):
            torch.cuda.synchronize()
        logger.info(f"time {(time.time() - tick)}")
        logger.info(f"error {torch.sum(Losses).item()}")

        if act_order:
            invperm = torch.argsort(perm)
            Q = Q[:, invperm]

        if is_transformers_imported() and isinstance(self.layer, transformers.Conv1D):
            Q = Q.t()
        # self.layer.weight.data = Q.reshape(self.layer.weight.shape).to(self.layer.weight.data.dtype)
        Q = Q.reshape(weight_shape).to(weight_dtype)
        if DEBUG:
            logger.info(f"{torch.sum((self.layer(self.inp1) - self.out1) ** 2)}")

        if scale == []:
            scale.append(self.quantizer.scale)
            zero.append(self.quantizer.zero)
        scale = torch.cat(scale, dim=1)
        zero = torch.cat(zero, dim=1)
        if "hpu" in self.device:
            scale = scale.to(self.device)
            zero = zero.to(self.device)
            Q = Q.to(self.device)
        return scale, zero, Q

    def free(self):
        """Free memory."""
        if DEBUG:
            self.inp1 = None
            self.out1 = None
        self.H = None
        self.Losses = None
        self.Trace = None
        torch.cuda.empty_cache()


class Quantizer(nn.Module):
    """Quantizer."""

    def __init__(self, shape=1):
        """Init Quantizer."""
        super(Quantizer, self).__init__()
        self.maxq = 0
        self.register_buffer("scale", torch.zeros(shape))
        self.register_buffer("zero", torch.zeros(shape))

    def configure(self, weight_config_this_layer, norm=2.4, grid=100, maxshrink=0.8, trits=False):
        """Configure the quantizer."""
        for k, v in weight_config_this_layer.items():
            setattr(self, k, v)
        # self.maxq = torch.tensor(2**self.bits - 1)
        self.maxq = 2**self.bits - 1
        self.scheme = "sym" if self.sym else "asym"
        self.double_quant_scheme = "sym" if self.double_quant_sym else "asym"
        self.norm = norm
        self.grid = grid
        self.maxshrink = maxshrink
        if trits:
            self.maxq = -1

    def find_params(self, x, weight=False):
        """Find scale and zero for weight."""
        dev = x.device
        # NF4 FP4
        if self.dtype != "int":
            from .utility import quant_tensor

            tmp = x.clone()  # tmp will be replaced after quant_tensor
            _, scale, zero = quant_tensor(
                tmp,
                dtype=self.dtype,
                bits=self.bits,
                group_size=self.group_size,
                scheme=self.scheme,
                quantile=1.0,
                return_int=True,
                full_range=False,
                double_quant=self.use_double_quant,
                double_quant_dtype=self.double_quant_dtype,
                double_quant_bits=self.double_quant_bits,
                double_quant_scheme=self.double_quant_scheme,
                double_quant_group_size=self.double_quant_group_size,
                double_quant_return_int=False,
            )
            self.scale = scale
            self.zero = torch.zeros_like(scale)
            return
        # INT
        shape = x.shape
        if self.perchannel:
            if weight:
                x = x.flatten(1)
            else:
                if len(shape) == 4:
                    x = x.permute([1, 0, 2, 3])
                    x = x.flatten(1)
                if len(shape) == 3:
                    x = x.reshape((-1, shape[-1])).t()
                if len(shape) == 2:
                    x = x.t()
        else:
            x = x.flatten().unsqueeze(0)

        tmp = torch.zeros(x.shape[0], device=dev)
        xmin = torch.minimum(x.min(1)[0], tmp)
        xmax = torch.maximum(x.max(1)[0], tmp)

        if self.sym:
            xmax = torch.maximum(torch.abs(xmin), xmax)
            tmp = xmin < 0
            if torch.any(tmp):
                xmin[tmp] = -xmax[tmp]
        tmp = (xmin == 0) & (xmax == 0)
        xmin[tmp] = -1
        xmax[tmp] = +1

        if self.maxq < 0:
            self.scale = xmax
            self.zero = xmin
        else:
            self.scale = (xmax - xmin) / self.maxq
            if self.sym:
                self.zero = torch.full_like(self.scale, (self.maxq + 1) / 2)
            else:
                self.zero = torch.round(-xmin / self.scale)

        if self.mse:
            best = torch.full([x.shape[0]], float("inf"), device=dev)
            for i in range(int(self.maxshrink * self.grid)):
                p = 1 - i / self.grid
                xmin1 = p * xmin
                xmax1 = p * xmax
                scale1 = (xmax1 - xmin1) / self.maxq
                zero1 = torch.round(-xmin1 / scale1) if not self.sym else self.zero
                q = self.quantize(x, scale1.unsqueeze(1), zero1.unsqueeze(1), self.maxq)
                q -= x
                q.abs_()
                q.pow_(self.norm)
                err = torch.sum(q, 1)
                tmp = err < best
                if torch.any(tmp):
                    best[tmp] = err[tmp]
                    self.scale[tmp] = scale1[tmp]
                    self.zero[tmp] = zero1[tmp]
        if not self.perchannel:
            if weight:
                tmp = shape[0]
            else:
                tmp = shape[1] if len(shape) != 3 else shape[2]
            self.scale = self.scale.repeat(tmp)
            self.zero = self.zero.repeat(tmp)

        if weight:
            shape = [-1] + [1] * (len(shape) - 1)
            self.scale = self.scale.reshape(shape)
            self.zero = self.zero.reshape(shape)

            if self.use_double_quant:
                # for INT
                from .utility import quant_tensor

                orig_scale_shape = self.scale.shape
                self.scale = self.scale.reshape(1, -1)
                quant_tensor(
                    self.scale,
                    dtype=self.double_quant_dtype,
                    bits=self.double_quant_bits,
                    group_size=self.double_quant_group_size,
                    scheme=self.double_quant_scheme,
                    quantile=1.0,
                    return_int=False,
                    full_range=False,
                )
                self.scale = self.scale.reshape(orig_scale_shape)
            return
        if len(shape) == 4:
            self.scale = self.scale.reshape((1, -1, 1, 1))
            self.zero = self.zero.reshape((1, -1, 1, 1))
        if len(shape) == 3:
            self.scale = self.scale.reshape((1, 1, -1))
            self.zero = self.zero.reshape((1, 1, -1))
        if len(shape) == 2:
            self.scale = self.scale.unsqueeze(0)
            self.zero = self.zero.unsqueeze(0)

    def quantize(self, x, scale, zero, maxq):
        """Do quantization."""
        if self.dtype != "int":
            from .utility import quantize_4bit

            tmp = x.clone()  # tmp will be replaced after quant_tensor
            return quantize_4bit(tmp, dtype=self.dtype, scale=scale)
        else:
            if maxq < 0:
                return (x > scale / 2).float() * scale + (x < zero / 2).float() * zero
            q = torch.clamp(torch.round(x / scale) + zero, 0, maxq)
            return scale * (q - zero)

    def ready(self):
        """Quantizer is ready.

        Returns:
            bool: True or False.
        """
        return torch.all(self.scale != 0)


from neural_compressor.torch.algorithms import Quantizer as INCQuantizer


class GPTQuantizer(INCQuantizer):
    """GPTQ Quantizer."""

    def __init__(self, quant_config={}):
        """Init a GPTQQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)

    @torch.no_grad()
    def prepare(
        self,
        model,
        nsamples=128,
        max_seq_length=2048,
        use_max_length=True,
        device=None,
        use_layer_wise=False,
        model_path=None,
        *args,
        **kwargs,
    ):
        """Run weight-only quantization with."""
        # TODO: unify weight_config keys, add docstring, and support default config
        assert isinstance(model, torch.nn.Module), "only support torch module"
        if use_layer_wise:  # pragma: no cover
            assert model_path is not None, "model_path should not be None when use layer wise mode"

        self.model_device = get_model_device(model)  # return model on the same device
        self.gptq_quantizer = RAWGPTQuantizer(
            model,
            weight_config=self.quant_config,
            nsamples=nsamples,
            use_max_length=use_max_length,
            max_seq_length=max_seq_length,
            device=device,
            use_layer_wise=use_layer_wise,
            model_path=model_path,
        )
        self.gptq_quantizer.prepare_for_calibration()
        return self.gptq_quantizer.model

    @torch.no_grad()
    def convert(self, model, *args, **kwargs):
        """Convert the prepared model to a quantized model.

        Args:
            model (torch.nn.Module): the prepared model

        Returns:
            The quantized model.
        """
        self.gptq_quantizer.model = model
        self.gptq_quantizer.remove_prepare_for_calibration()

        q_model, gptq_config = self.gptq_quantizer.execute_quantization()
        if not self.gptq_quantizer.use_layer_wise:
            q_model = q_model.to(self.model_device)
        q_model.gptq_config = gptq_config
        logger.info("GPTQ quantizing done.")
        return q_model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json

import numpy as np
import scipy

tasks = ["winogrande", "hellaswag", "piqa", "lambada_openai"]


def ztest(ref_mean=0.0, ref_stderr=1.0, test_mean=0.0, test_stderr=0.0):
    z_score = (test_mean - ref_mean) / np.sqrt(ref_stderr**2 + test_stderr**2)
    p_value = 1.0 + scipy.special.erf(-np.abs(z_score) / np.sqrt(2))
    return p_value


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Regression detection using Z-Test. We assume we have mean and SEM of golden run arranged in json and test results json and we compare the results to see if degregation occurred.",
    )
    parser.add_argument(
        "--hp_dtype",
        type=str,
        help="Data type of the high precision test",
        default=None,
    )
    parser.add_argument(
        "--lp_dtype",
        type=str,
        help="Data type of the low precision test",
        default=None,
    )
    parser.add_argument(
        "--golden_metrics",
        type=str,
        help="Path to json that includes mean, SEM and diff golden metrics of bf16 and fp8 precision.",
        default=None,
    )
    parser.add_argument(
        "--test_metrics_lp",
        type=str,
        help="Path to json that includes mean, SEM and diff test metrics of lp precision.",
        default=None,
    )
    parser.add_argument(
        "--test_metrics_hp",
        type=str,
        help="Path to json that includes mean, SEM and diff test metrics of high precision.",
        default=None,
    )
    parser.add_argument("--quantization_mode", type=str, help="quantization mode", default=None)
    args = parser.parse_args()
    mode = args.quantization_mode
    hp_dtype = args.hp_dtype
    lp_dtype = args.lp_dtype
    golden_metrics_path = args.golden_metrics
    test_metrics_lp_path = args.test_metrics_lp
    test_metrics_hp_path = args.test_metrics_hp
    if golden_metrics_path is None or test_metrics_hp_path is None or test_metrics_lp_path is None:
        print("Please provide golden_metrics, test_metrics_hp_path and test_metrics_lp_path json paths")
        exit(1)

    with open(golden_metrics_path, "r") as f:
        golden_metrics_json = json.load(f)

    with open(test_metrics_lp_path, "r") as f:
        test_metrics_lp_json = json.load(f)
        test_metrics_lp_json = test_metrics_lp_json["results"]

    with open(test_metrics_hp_path, "r") as f:
        test_metrics_hp_json = json.load(f)
        test_metrics_hp_json = test_metrics_hp_json["results"]

    regressions = []
    for task in tasks:
        # The two-sample z-test comparing the golden and under-test high-precision configuration
        ref_mean_hp = golden_metrics_json[hp_dtype][task]["mean"]
        ref_stderr_hp = golden_metrics_json[hp_dtype][task]["sem"]
        test_mean_hp = test_metrics_hp_json[task]["acc"]
        test_stderr_hp = test_metrics_hp_json[task]["acc_stderr"]
        p_hp_value = ztest(ref_mean_hp, ref_stderr_hp, test_mean_hp, test_stderr_hp)
        print(f"Z-Test high precision p-value={p_hp_value*100:.2f}%  in {task} task")
        if p_hp_value < 0.05:
            regressions.append(f"Z-Test high precision p-value is less than 0.05 in {task} task.")

        # The two-sample z-test comparing the golden and under-test low-precision configuration
        if mode is not None:
            ref_mean_lp = golden_metrics_json[lp_dtype][mode][task]["mean"]
            ref_stderr_lp = golden_metrics_json[lp_dtype][mode][task]["sem"]
        else:
            ref_mean_lp = golden_metrics_json[lp_dtype][task]["mean"]
            ref_stderr_lp = golden_metrics_json[lp_dtype][task]["sem"]
        test_mean_lp = test_metrics_lp_json[task]["acc"]
        test_stderr_lp = test_metrics_lp_json[task]["acc_stderr"]
        p_lp_value = ztest(ref_mean_lp, ref_stderr_lp, test_mean_lp, test_stderr_lp)
        print(f"Z-Test low precision p-value={p_lp_value*100:.2f}% in {task} task")
        if p_lp_value < 0.05:
            regressions.append(f"Z-Test low precision p-value is less than 0.05 in {task} task.")

        # The single-sample z-test comparing the golden and under-test degradation of low-precision configuration
        if mode is not None:
            ref_mean_diff = golden_metrics_json[lp_dtype][mode][task]["mean_diff"]
            ref_stderr_diff = golden_metrics_json[lp_dtype][mode][task]["sem_diff"]
        else:
            ref_mean_diff = golden_metrics_json[lp_dtype][task]["mean_diff"]
            ref_stderr_diff = golden_metrics_json[lp_dtype][task]["sem_diff"]
        test_mean_diff = test_mean_lp - test_mean_hp
        p_diff_value = ztest(ref_mean_diff, ref_stderr_diff, test_mean_diff, ref_stderr_diff)
        print(f"Z-Test low precision diff p-value={p_diff_value*100:.2f}% in {task} task")
        if p_diff_value < 0.05:
            regressions.append(f"Z-Test low precision diff p-value is less than 0.05 in {task} task.")

    if len(regressions) == 0:
        print("No regressions were detected!")
    else:
        print("Regressions were detected!")
        for regression in regressions:
            print(regression)
